// 
// Author:
//       smdn <smdn@mail.invisiblefulmoon.net>
// 
// Copyright (c) 2010 smdn
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Diagnostics;
using System.IO;
using System.Reflection;

namespace Smdn.Net.MessageAccessProtocols {
  internal static class ConnectionTrace {
    [Conditional("TRACE")]
    public static void Log(ConnectionBase connection, Exception ex)
    {
      connection.traceSource.TraceEvent(TraceEventType.Error, connection.Id, "{0:o} CID:{1} {2}",
                                        DateTime.Now,
                                        connection.Id,
                                        ex);
    }

    [Conditional("TRACE")]
    public static void Log(ConnectionBase connection, string format, params object[] args)
    {
      connection.traceSource.TraceEvent(TraceEventType.Information, connection.Id, "{0:o} CID:{1} {2}",
                                        DateTime.Now,
                                        connection.Id,
                                        string.Format(format, args));
    }

    [Conditional("TRACE")]
    private static void LogReceived(ConnectionBase connection, byte[] received, int start, int count)
    {
      var recv = NetworkTransferEncoding.Transfer8Bit.GetString(received, start, count);

      connection.traceSource.TraceEvent(TraceEventType.Verbose, connection.Id, "{0:o} CID:{1} [{2, 17}] S: {3}",
                                        DateTime.Now,
                                        connection.Id,
                                        connection.RemoteEndPoint,
                                        recv.TrimEnd());
    }

    [Conditional("TRACE")]
    private static void LogSent(ConnectionBase connection, byte[] sent, int start, int count)
    {
      var snt = NetworkTransferEncoding.Transfer8Bit.GetString(sent, start, count);

      connection.traceSource.TraceEvent(TraceEventType.Verbose, connection.Id, "{0:o} CID:{1} [{2, 17}] C: {3}",
                                        DateTime.Now,
                                        connection.Id,
                                        connection.LocalEndPoint,
                                        snt.TrimEnd());
    }

    internal static Stream CreateTracingStream(ConnectionBase connection, Stream stream)
    {
#if TRACE
      // FIXME: mono bug?
      if (((int)connection.traceSource.Switch.Level & (int)TraceEventType.Verbose) != 0)
        return new TracingStream(connection, stream);
      else
#endif
        return stream;
    }

#if TRACE
    internal class TracingStream : InterruptStream {
      public TracingStream(ConnectionBase connection, Stream innerStream)
        : base(innerStream)
      {
        this.connection = connection;
      }

      protected override void OnWritten(byte[] src, int offset, int count)
      {
        ConnectionTrace.LogSent(connection, src, offset, count);
      }

      protected override int OnRead(byte[] dest, int offset, int count)
      {
        ConnectionTrace.LogReceived(connection, dest, offset, count);

        return count;
      }

      private ConnectionBase connection;
    }
#endif
  }
}
