// 
// Author:
//       smdn <smdn@mail.invisiblefulmoon.net>
// 
// Copyright (c) 2010 smdn
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.Net;

using Smdn.Collections;
using Smdn.Net.Pop3.Client.Transaction;
using Smdn.Net.Pop3.Protocol;
using Smdn.Net.MessageAccessProtocols;
using Smdn.Security.Authentication.Sasl.Client;

namespace Smdn.Net.Pop3.Client.Session {
  public static class PopSessionCreator {
    public static PopSession CreateSession(IPopSessionProfile profile,
                                            UpgradeConnectionStreamCallback createSslStreamCallback)
    {
      PopSession session;

      var result = CreateSession(profile, createSslStreamCallback, out session);

      if (result.Succeeded)
        return session;
      else
        throw new PopException("create session failed: " + result.ResultText);
    }

    public static PopTransactionResult CreateSession(IPopSessionProfile profile,
                                                     UpgradeConnectionStreamCallback createSslStreamCallback,
                                                     out PopSession session)
    {
      if (profile == null)
        throw new ArgumentNullException("profile");

      var authority = profile.Authority;
      var securePort = string.Equals(authority.Scheme, PopUri.UriSchemePops, StringComparison.InvariantCultureIgnoreCase);

      if (securePort && createSslStreamCallback == null)
        throw new ArgumentNullException("createSslStreamCallback");

      PopTransactionResult result;
      session = null;

      session = new PopSession(authority.Host,
                               authority.Port,
                               profile.Timeout,
                               securePort
                                 ? createSslStreamCallback
                                 : null);

      session.HandlesIncapableAsException = false;
      session.TransactionTimeout  = profile.Timeout;
      session.SendTimeout         = profile.SendTimeout;
      session.ReceiveTimeout      = profile.ReceiveTimeout;

      // try querying server capability (ignore error; POP3 Extension Mechanism might not supported)
      session.Capa();

      if (profile.UseTlsIfAvailable && session.ServerCapabilities.IsCapable(PopCapability.Stls) && !session.IsSecureConnection) {
        var r = session.Stls(createSslStreamCallback, false);

        if (r.Failed)
          throw new PopUpgradeConnectionException(r.ResultText);

        // try re-querying server capability (ignore error; POP3 Extension Mechanism might not supported)
        session.Capa();
      }

      result = Authenticate(session, profile);

      if (result == null) {
        throw new PopException("appropriate authentication mechanism not found");
      }
      else if (result.Failed) {
        try {
          try {
            session.Disconnect(false);
          }
          catch (PopConnectionException) {
            // ignore
          }
        }
        finally {
          session = null;
        }
      }

      return result;
    }

    private static PopTransactionResult Authenticate(PopSession session, IPopSessionProfile profile)
    {
      var authority = profile.Authority;
      var username = PopStyleUriParser.GetUser(authority);
      var authMechanism = PopStyleUriParser.GetAuthType(authority);

      /*
       * http://tools.ietf.org/html/rfc2384
       * 4. POP User Name and Authentication Mechanism
       * 
       *    The string ";AUTH=*" indicates that the client SHOULD select an
       *    appropriate authentication mechanism.  It MAY use any mechanism
       *    supported by the POP server.
       * 
       *    If a user name is included with no authentication mechanism, then
       *    ";AUTH=*" is assumed.
       */
      var canFallback = false;

      if (authMechanism == null) {
        if (string.IsNullOrEmpty(username)) {
          authMechanism = PopAuthenticationMechanism.Anonymous;
          canFallback = true;
        }
        else {
          authMechanism = PopAuthenticationMechanism.SelectAppropriate;
        }
      }

      if (authMechanism == PopAuthenticationMechanism.SelectAppropriate)
        return AuthenticateWithAppropriateMechanism(session,
                                                    profile.Credentials,
                                                    username,
                                                    GetUsingSaslMechanisms(profile.UsingSaslMechanisms ?? new string[0]));
      else if (authMechanism == PopAuthenticationMechanism.Anonymous)
        return AuthenticateAsAnonymous(session,
                                       profile.AnonymousCredential,
                                       canFallback);
      else
        return AuthenticateWithSuppliedMechanism(session,
                                                 profile.Credentials,
                                                 username,
                                                 authMechanism);
    }

    private static IEnumerable<PopAuthenticationMechanism> GetUsingSaslMechanisms(string[] usingSaslMechanisms)
    {
      var availableMechanisms = Smdn.Security.Authentication.Sasl.Client.SaslClientMechanism.GetAvailableMechanisms();

      foreach (var mechanism in usingSaslMechanisms) {
        if (availableMechanisms.Exists(delegate(string val) {
          return string.Equals(val, mechanism, StringComparison.InvariantCultureIgnoreCase);
        }))
          yield return PopAuthenticationMechanism.GetKnownOrCreate(mechanism);
      }
    }

    private static PopTransactionResult AuthenticateWithAppropriateMechanism(PopSession session, ICredentialsByHost credentials, string username, IEnumerable<PopAuthenticationMechanism> usingSaslMechanisms)
    {
      PopTransactionResult result = null;

      foreach (var mechanism in usingSaslMechanisms) {
        if (session.ServerCapabilities.IsCapable(mechanism)) {
          result = session.Auth(credentials, username, mechanism);

          if (result.Succeeded)
            break;
        }
      }

      if (result == null || result.Failed) {
        if (session.ApopAvailable)
          result = session.Apop(credentials, username);
      }

      if (result == null || result.Failed)
        result = session.Login(credentials, username);

      return result;
    }

    private static PopTransactionResult AuthenticateWithSuppliedMechanism(PopSession session, ICredentialsByHost credentials, string username, PopAuthenticationMechanism authMechanism)
    {
      /*
       * http://tools.ietf.org/html/rfc2384
       * 4. POP User Name and Authentication Mechanism
       * 
       *    An authentication mechanism can be expressed by adding ";AUTH=<enc-
       *    auth-type>" to the end of the user name.  If the authentication
       *    mechanism name is not preceded by a "+", it is a SASL POP [SASL]
       *    mechanism.  If it is preceded by a "+", it is either "APOP" or an
       *    extension mechanism.
       * 
       *    When an <enc-auth-type> is specified, the client SHOULD request
       *    appropriate credentials from that mechanism and use the "AUTH",
       *    "APOP", or extension command instead of the "USER" command.  If no
       *    user name is specified, one SHOULD be obtained from the mechanism or
       *    requested from the user as appropriate.
       * 
       * 
       *    If an <enc-auth-type> other than ";AUTH=*" is specified, the client
       *    SHOULD NOT use a different mechanism without explicit user
       *    permission.
       */
      return session.Auth(credentials, username, authMechanism);
    }

    private static PopTransactionResult AuthenticateAsAnonymous(PopSession session, NetworkCredential anonymousCredential, bool canFallback)
    {
      PopTransactionResult result = null;

      if (session.ServerCapabilities.IsCapable(PopAuthenticationMechanism.Anonymous))
        // try AUTH ANONYUMOUS
        result = session.Auth(anonymousCredential, PopAuthenticationMechanism.Anonymous);

      if (result == null || (result.Failed && canFallback))
        // try anonymous LOGIN
        result = session.Login(anonymousCredential);

      return result;
    }
  }
}
