// 
// Author:
//       smdn <smdn@mail.invisiblefulmoon.net>
// 
// Copyright (c) 2010 smdn
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;

using Smdn.Net.Pop3.Protocol;
using Smdn.Net.Pop3.Protocol.Client;

namespace Smdn.Net.Pop3.Client.Transaction {
  public abstract class PopTransactionBase : PopTransactionBase<PopTransactionResult> {
    protected PopTransactionBase(PopConnection connection)
      : base(connection)
    {
    }
  }

  public abstract class PopTransactionBase<TResult> : PopTransactionBase<PopTransactionRequest, TResult>
    where TResult : PopTransactionResult, new()
  {
    protected PopTransactionBase(PopConnection connection)
      : base(connection, new PopTransactionRequest())
    {
    }
  }

  public abstract class PopTransactionBase<TRequest, TResult> : IPopTransaction, IDisposable
    where TRequest : PopTransactionRequest
    where TResult : PopTransactionResult, new()
  {
    protected delegate void ProcessTransactionDelegate();

    public PopConnection Connection {
      get
      {
        CheckDisposed();
        return connection;
      }
    }

    public TRequest Request {
      get
      {
        CheckDisposed();
        return request;
      }
    }

    PopTransactionRequest IPopTransaction.Request {
      get { return this.Request; }
    }

    public TResult Result {
      get
      {
        CheckDisposed();
        return result;
      }
    }

    PopTransactionResult IPopTransaction.Result {
      get { return this.Result; }
    }

    public PopTransactionState State {
      get
      {
        CheckDisposed();
        return state;
      }
    }

    public bool IsResponseMultiline {
      get; protected set;
    }

    protected PopStatusResponse StatusResponse {
      get; private set;
    }

    protected PopTransactionBase(PopConnection connection, TRequest request)
    {
      if (connection == null)
        throw new ArgumentNullException("connection");
      if (request == null)
        throw new ArgumentNullException("request");

      this.connection = connection;
      this.request = request;
      this.result = null;
      this.receivedResponses = new Queue<PopResponse>();
      this.IsResponseMultiline = false;
    }

    public virtual void Dispose()
    {
      if (disposed)
        return;

      // nothing to do in this class
      connection = null;
      request = null;
      result = null;
      receivedResponses = null;

      disposed = true;
    }

    private PopConnection connection;
    private TRequest request;
    private TResult result;
    private bool disposed = false;

#region "processing transaction / state transition"
    void IPopTransaction.Start()
    {
      CheckDisposed();

      if (state == PopTransactionState.Finished)
        throw new InvalidOperationException("transaction already finished");

      currentProcess = Reset();

      state = PopTransactionState.Processing;

      connection.HandleResponseAsMultiline = false;
    }

    protected abstract ProcessTransactionDelegate Reset();

    void IPopTransaction.Process()
    {
      CheckDisposed();
      RejectFinished();

      try {
        if (currentProcess != null)
          currentProcess();
      }
      catch (Exception ex) {
        Finish(CreateExceptionResult(ex));
        Trace.Log(ex);
      }
    }

    /// <remarks>do not call directly</remarks>
    protected void ProcessReceiveResponse()
    {
      var response = Receive();

      if (response == null)
        return;

      try {
        OnResponseReceived(response);
      }
      catch (PopMalformedTextException ex) {
        if (response is PopStatusResponse) {
          // error in response data, abort transaction
          throw ex;
        }
        else {
          // error in response data, continue processing
          Trace.Log(ex);
        }
      }
    }

    protected void Transit(ProcessTransactionDelegate process)
    {
      if (state == PopTransactionState.Finished)
        throw new InvalidOperationException("transaction already finished");
      if (process == null)
        throw new ArgumentNullException("process");

      currentProcess = process;
    }

    private TResult CreateExceptionResult(Exception exception)
    {
      if (exception is TimeoutException)
        return new TResult() {Exception = exception, Code = PopTransactionResultCode.SocketTimeout, Description = "timed out"};
      else if (exception is PopUpgradeConnectionException)
        return new TResult() {Exception = exception, Code = PopTransactionResultCode.UpgradeError, Description = "upgrade connection failed"};
      else
        return new TResult() {Exception = exception, Code = PopTransactionResultCode.InternalError, Description = "internal error"};
    }

    protected void FinishError(PopStatusResponse status)
    {
      switch (status.Status) {
        case PopStatusIndicator.Negative:
          Finish(new TResult() {Code = PopTransactionResultCode.Error, ResponseText = status.Text});
          break;
        default: // positive or unexpected
          throw new InvalidOperationException("status is not -ERR");
      }
    }

    protected void FinishError(PopTransactionResultCode code, string description)
    {
      if ((int)code < 300)
        throw new ArgumentException("status is not error", "status");

      Finish(new TResult() {Code = code, Description = description});
    }

    protected void FinishOk(PopStatusResponse status)
    {
      switch (status.Status) {
        case PopStatusIndicator.Positive:
          Finish(new TResult() {Code = PopTransactionResultCode.Ok, ResponseText = status.Text});
          break;
        default: // negative or unexpected
          throw new InvalidOperationException("status is not +OK");
      }
    }

    protected internal void Finish(TResult result)
    {
      RejectFinished();

      this.result = result;
      this.result.ReceivedResponses = receivedResponses;
      this.result.StatusResponse = StatusResponse;

      currentProcess = null;

      state = PopTransactionState.Finished;
    }

    private void RejectFinished()
    {
      if (state == PopTransactionState.Finished)
        throw new InvalidOperationException("transaction already finished");
    }

    private PopTransactionState state = PopTransactionState.NotStarted;
    private ProcessTransactionDelegate currentProcess;
#endregion

#region "sending / receiving"
    protected PopResponse Receive()
    {
      RejectFinished();

      try {
        var resp = Connection.TryReceiveResponse();

        if (resp != null)
          receivedResponses.Enqueue(resp);

        return resp;
      }
      catch (PopMalformedResponseException ex) {
        // protocol error, abort transaction
        throw ex;
      }
    }

    protected void SendCommand(string command, ProcessTransactionDelegate nextProcess, params string[] arguments)
    {
      RejectFinished();

      Connection.SendCommand(new PopCommand(command, arguments));

      Transit(nextProcess);
    }

    protected void SendContinuation(params string[] arguments)
    {
      RejectFinished();

      Connection.SendCommand(new PopCommand(null, arguments));
    }

    private Queue<PopResponse> receivedResponses;
#endregion

#region "response processing"
    private void OnResponseReceived(PopResponse response)
    {
      if (response is PopFollowingResponse)
        OnFollowingResponseReceived(response as PopFollowingResponse);
      else if (response is PopStatusResponse)
        OnStatusResponseReceived(response as PopStatusResponse);
      else if (response is PopTerminationResponse)
        OnTerminationResponse(response as PopTerminationResponse);
      else if (response is PopContinuationRequest)
        OnContinuationRequestReceived(response as PopContinuationRequest);
    }

    protected virtual void OnFollowingResponseReceived(PopFollowingResponse following)
    {
    }

    protected virtual void OnStatusResponseReceived(PopStatusResponse status)
    {
      if (IsResponseMultiline) {
        if (status.Status == PopStatusIndicator.Positive) {
          if (StatusResponse == null)
            Connection.HandleResponseAsMultiline = true;
          StatusResponse = status;
        }
        else {
          StatusResponse = status;

          FinishError(status);
        }
      }
      else {
        StatusResponse = status;

        if (status.Status == PopStatusIndicator.Positive)
          FinishOk(status);
        else
          FinishError(status);
      }
    }

    protected virtual void OnTerminationResponse(PopTerminationResponse termination)
    {
      if (IsResponseMultiline)
        FinishOk(StatusResponse);
    }

    protected virtual void OnContinuationRequestReceived(PopContinuationRequest continuationRequest)
    {
    }
#endregion

    public override string ToString()
    {
      return string.Format("{{Connection={0}, State={1}, Request={2}, Result={3}}}", connection, state, request, result);
    }

    private void CheckDisposed()
    {
      if (disposed)
        throw new ObjectDisposedException(GetType().FullName);
    }
  }
}
