// 
// Author:
//       smdn <smdn@mail.invisiblefulmoon.net>
// 
// Copyright (c) 2010 smdn
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.IO;
using System.Text;

using Smdn.Collections;
using Smdn.IO;
using Smdn.Net.Pop3.Client.Session;

namespace Smdn.Net.Pop3.Client {
  public class PopMessageInfo : IPopMessageNumber {
    public long MessageNumber {
      get { return scanList.MessageNumber; }
    }

    public long Length {
      get { return scanList.SizeInOctets; }
    }

    public string UniqueId {
      get { return EnsureUniqueIdListed().UniqueId; }
    }

    public bool IsMarkedAsDeleted {
      get; private set;
    }

    internal PopMessageInfo(PopClient client, PopScanListing scanList)
    {
      this.client = client;
      this.scanList = scanList;
      this.IsMarkedAsDeleted = false;
    }

    private PopClient client;
    private PopScanListing scanList;
    private PopUniqueIdListing uniqueIdList;

    /*
     * operation
     */
    public void MarkAsDeleted()
    {
      PopClient.ThrowIfError(client.Session.Dele(MessageNumber));

      IsMarkedAsDeleted = true;
    }

    internal void Reset()
    {
      IsMarkedAsDeleted = false;
    }

    private PopUniqueIdListing EnsureUniqueIdListed()
    {
      if (uniqueIdList != null)
        return uniqueIdList;

      PopClient.ThrowIfError(client.Session.Uidl(MessageNumber, out uniqueIdList));

      return uniqueIdList;
    }

    private const int entireMessage = -1;

    public Stream OpenRead()
    {
      return OpenRead(entireMessage);
    }

    /// <param name="maxLines">
    /// A maximum line count to retrieve.
    /// <list type="number">
    ///   <listheader><term>value</term><description>behavior</description></listheader>
    ///   <item><term>-1</term><description>entire message</description></item>
    ///   <item><term>0</term><description>message without body</description></item>
    ///   <item><term>1 or greater</term><description>message with specified lines of body</description></item>
    /// </list>
    /// </param>
    public Stream OpenRead(int maxLines)
    {
      CheckMarkedAsDeleted();

      if (maxLines < -1)
        throw new ArgumentOutOfRangeException("maxLines", "must be -1, 0 or positive number");

      Stream messageStream;

      if (maxLines == entireMessage)
        PopClient.ThrowIfError(client.Session.Retr(MessageNumber, out messageStream));
      else
        PopClient.ThrowIfError(client.Session.Top(MessageNumber, maxLines, out messageStream));

      if (client.DeleteAfterRetrieve)
        MarkAsDeleted();

      return messageStream;
    }

    public TOutput ReadAs<TOutput>(Converter<Stream, TOutput> converter)
    {
      return ReadAs(entireMessage, converter);
    }

    public TOutput ReadAs<TOutput>(int maxLines, Converter<Stream, TOutput> converter)
    {
      if (converter == null)
        throw new ArgumentNullException("converter");

      using (var stream = OpenRead(maxLines)) {
        return converter(stream);
      }
    }

    public StreamReader OpenText()
    {
      return OpenText(Encoding.Default, entireMessage);
    }

    public StreamReader OpenText(Encoding encoding)
    {
      return OpenText(encoding, entireMessage);
    }

    public StreamReader OpenText(int maxLines)
    {
      return OpenText(Encoding.Default, maxLines);
    }

    public StreamReader OpenText(Encoding encoding, int maxLines)
    {
      if (encoding == null)
        throw new ArgumentNullException("encoding");

      return new StreamReader(OpenRead(maxLines), encoding, false);
    }

    public byte[] ReadAllBytes()
    {
      return ReadAllBytes(entireMessage);
    }

    public byte[] ReadAllBytes(int maxLines)
    {
      using (var stream = OpenRead(maxLines)) {
        return stream.ReadToEnd();
      }
    }

    public string[] ReadAllLines(Encoding encoding)
    {
      return ReadAllLines(encoding, entireMessage);
    }

    public string[] ReadAllLines(Encoding encoding, int maxLines)
    {
      return ReadLines(encoding, maxLines).ToArray();
    }

    public IEnumerable<string> ReadLines(Encoding encoding)
    {
      return ReadLines(encoding, entireMessage);
    }

    public IEnumerable<string> ReadLines(Encoding encoding, int maxLines)
    {
      if (encoding == null)
        throw new ArgumentNullException("encoding");

      using (var stream = OpenRead(maxLines)) {
        var reader = new StreamReader(stream, encoding);

        for (;;) {
          var l = reader.ReadLine();

          if (l == null)
            yield break;
          else
            yield return l;
        }
      }
    }

    public string ReadAllText(Encoding encoding)
    {
      return ReadAllText(encoding, entireMessage);
    }

    public string ReadAllText(Encoding encoding, int maxLines)
    {
      if (encoding == null)
        throw new ArgumentNullException("encoding");

      using (var stream = OpenRead(maxLines)) {
        return (new StreamReader(stream, encoding)).ReadToEnd();
      }
    }

    private void CheckMarkedAsDeleted()
    {
      if (IsMarkedAsDeleted)
        throw new PopProtocolViolationException("The message has been marked as deleted.");
    }

    public override string ToString()
    {
      return string.Format("{{PopMessageInfo: Authority='{0}', MessageNumber={1}, Length={2}}}",
                           client.Profile.Authority,
                           MessageNumber,
                           Length);
    }
  }
}
