using System;
using System.IO;
using System.Net;
using System.Threading;
using NUnit.Framework;

using Smdn.Net.MessageAccessProtocols;

using Smdn.Net.Pop3.Protocol;
using Smdn.Net.Pop3.Protocol.Client;

namespace Smdn.Net.Pop3.Client.Session {
  [TestFixture]
  public class PopSessionTests {
    class NullCredential : ICredentialsByHost {
      public NetworkCredential GetCredential(string host, int port, string authType)
      {
        return null;
      }
    }

    [SetUp]
    public void Setup()
    {
      server = new PopPseudoServer();
      server.Start();

      host = server.ServerEndPoint.Address.ToString();
      port = server.ServerEndPoint.Port;

      credential = new NetworkCredential(username, password, host);
      authority = new Uri(string.Format("{0}://{1}@{2}:{3}",
                                  PopUri.UriSchemePop,
                                  username,
                                  host,
                                  port));
    }

    [TearDown]
    public void TearDown()
    {
      server.Stop();
    }

    [Test]
    public void TestConnect()
    {
      server.EnqueueResponse("+OK PopPseudoServer ready\r\n");

      using (var session = new PopSession(host, port)) {
        Assert.IsFalse(session.ApopAvailable);
        Assert.IsTrue(string.IsNullOrEmpty(session.Timestamp));
        Assert.AreEqual(PopSessionState.Authorization, session.State);
        Assert.AreEqual(new Uri(string.Format("pop://{0}:{1}/", host, port)), session.Authority);
        Assert.AreEqual(0, session.ServerCapabilities.Count);
      }
    }

    [Test, ExpectedException(typeof(Smdn.Net.Pop3.Protocol.PopConnectionException))]
    public void TestConnectTransactionTimeout()
    {
      server.Stop();

      using (var session = new PopSession(host, port, 500)) {
      }
    }

    [Test, ExpectedException(typeof(Smdn.Net.Pop3.Protocol.PopConnectionException))]
    public void TestConnectErr()
    {
      server.EnqueueResponse("-ERR try again\r\n");

      using (var session = new PopSession(host, port)) {
      }
    }

    [Test, ExpectedException(typeof(Smdn.Net.Pop3.Protocol.PopConnectionException))]
    public void TestConnectInvalidResponse()
    {
      server.EnqueueResponse("* OK pseudo IMAP server ready\r\n");

      using (var session = new PopSession(host, port)) {
      }
    }

    [Test]
    public void TestConnectSsl()
    {
      server.EnqueueResponse("+OK PopPseudoServer ready\r\n");

      var streamUpgraded = false;

      using (var session = new PopSession(host, port, delegate(ConnectionBase connection, Stream baseStream) {
        streamUpgraded = true;
        return baseStream; // TODO: return SslStream
      })) {
        Assert.IsFalse(session.ApopAvailable);
        Assert.IsTrue(string.IsNullOrEmpty(session.Timestamp));
        Assert.AreEqual(PopSessionState.Authorization, session.State);
        Assert.AreEqual(new Uri(string.Format("pops://{0}:{1}/", host, port)), session.Authority);
        Assert.AreEqual(0, session.ServerCapabilities.Count);
        Assert.IsTrue(streamUpgraded, "stream upgraded");
        Assert.IsTrue(session.IsSecureConnection);
      }
    }

    [Test, ExpectedException(typeof(PopUpgradeConnectionException))]
    public void TestConnectSslExceptionWhileUpgrading()
    {
      server.EnqueueResponse("+OK PopPseudoServer ready\r\n");

      using (var session = new PopSession(host, port, delegate(ConnectionBase connection, Stream baseStream) {
        throw new System.Security.Authentication.AuthenticationException();
      })) {
        Assert.Fail("connected");
      }
    }

    [Test]
    public void TestDisconnect()
    {
      server.EnqueueResponse("+OK PopPseudoServer ready\r\n");

      var session = new PopSession(host, port);

      session.Disconnect(false);

      Assert.IsTrue(session.IsDisposed);
    }

    [Test]
    public void TestGreetingWithTimestamp()
    {
      server.EnqueueResponse("+OK POP3 server ready <1896.697170952@dbc.mtview.ca.us>\r\n");

      using (var session = new PopSession(host, port)) {
        Assert.IsTrue(session.ApopAvailable);
        Assert.AreEqual("<1896.697170952@dbc.mtview.ca.us>", session.Timestamp);
        Assert.AreEqual(PopSessionState.Authorization, session.State);
        Assert.AreEqual(new Uri(string.Format("pop://{0}:{1}/", host, port)), session.Authority);
        Assert.AreEqual(0, session.ServerCapabilities.Count);
      }
    }

    private PopSession Connect(string timestamp)
    {
      if (string.IsNullOrEmpty(timestamp))
        server.EnqueueResponse("+OK POP3 server ready\r\n");
      else
        server.EnqueueResponse(string.Format("+OK POP3 server ready {0}\r\n", timestamp));

      var session = new PopSession(host, port);

      Assert.AreEqual(PopSessionState.Authorization, session.State);
      Assert.AreEqual(authority, session.Authority);

      return session;
    }

    [Test, ExpectedException(typeof(NotSupportedException))]
    public void TestChangeReadOnlyPropertyServerCapabilities()
    {
      using (var session = Connect(null)) {
        session.ServerCapabilities.Add(PopCapability.User);
      }
    }

    [Test]
    public void TestTransactionTimeoutWhileTransactionProceeding()
    {
      using (var session = Connect(null)) {
        session.TransactionTimeout = 250;
        session.SendTimeout = System.Threading.Timeout.Infinite;
        session.ReceiveTimeout = System.Threading.Timeout.Infinite;

        // not respond CAPA response termination
        server.EnqueueResponse("+OK List of capabilities follows\r\n");

        try {
          session.Capa();
          Assert.Fail("TimeoutException not thrown");
        }
        catch (TimeoutException) {
          Assert.AreEqual(PopSessionState.NotConnected, session.State);
        }
      }
    }

    [Test]
    public void TestSocketTimeoutWhileTransactionProceeding()
    {
      using (var session = Connect(null)) {
        session.TransactionTimeout = System.Threading.Timeout.Infinite;
        session.SendTimeout = 250;
        session.ReceiveTimeout = 250;

        // not respond CAPA
        //server.EnqueueResponse("+OK List of capabilities follows\r\n.\r\n");

        try {
          session.Capa();
          Assert.Fail("TimeoutException not thrown");
        }
        catch (TimeoutException) {
          Assert.AreEqual(PopSessionState.NotConnected, session.State);
        }
      }
    }

    [Test]
    public void TestInternalErrorWhileTransactionProceeding()
    {
      using (var session = Connect(null)) {
        server.EnqueueResponse("+OK List of capabilities follows\r\n" +
                               "\r\n" + // invalid response
                               ".\r\n");

        try {
          session.Capa();
          Assert.Fail("PopException not thrown");
        }
        catch (PopException ex) {
          Assert.IsNotNull(ex.InnerException);
          Assert.AreEqual(PopSessionState.NotConnected, session.State);
        }
      }
    }

    [Test, ExpectedException(typeof(InvalidOperationException))]
    public void TestBeginNewTransactionWhileTransactionProceeding()
    {
      using (var session = Connect(null)) {
        ThreadPool.QueueUserWorkItem(delegate(object state) {
          (state as PopSession).Capa();
        }, session);

        Thread.Sleep(250);

        Assert.IsTrue(session.IsTransactionProceeding);

        session.Capa();
      }
    }

    [Test]
    public void TestQuitInAuthorizationState()
    {
      using (var session = Connect(null)) {
        server.EnqueueResponse("+OK dewey POP3 server signing off\r\n");

        Assert.IsTrue((bool)session.Quit());

        StringAssert.AreEqualIgnoringCase("QUIT\r\n",
                                          server.DequeueRequest());

        Assert.AreEqual(PopSessionState.NotConnected, session.State);
        Assert.IsNull(session.Authority);
      }
    }

    [Test]
    public void TestQuitAfterDisconnected()
    {
      using (var session = Connect(null)) {
        server.EnqueueResponse("+OK dewey POP3 server signing off\r\n");

        Assert.IsTrue((bool)session.Quit());

        StringAssert.AreEqualIgnoringCase("QUIT\r\n",
                                          server.DequeueRequest());

        Assert.AreEqual(PopSessionState.NotConnected, session.State);

        Assert.IsTrue((bool)session.Quit());

        Assert.AreEqual(PopSessionState.NotConnected, session.State);
      }
    }

    [Test]
    public void TestUser()
    {
      using (var session = Connect(null)) {
        server.EnqueueResponse("-ERR sorry, no mailbox for frated here\r\n");

        Assert.IsFalse((bool)session.User("frated"));
        Assert.AreEqual(PopSessionState.Authorization, session.State);
        Assert.AreEqual(authority, session.Authority);

        StringAssert.AreEqualIgnoringCase("USER frated\r\n",
                                          server.DequeueRequest());

        server.EnqueueResponse("+OK mrose is a real hoopy frood\r\n");

        Assert.IsTrue((bool)session.User("mrose"));

        StringAssert.AreEqualIgnoringCase("USER mrose\r\n",
                                          server.DequeueRequest());

        Assert.AreEqual(PopSessionState.Authorization, session.State);
        Assert.AreEqual(authority, session.Authority);
      }
    }

    [Test]
    public void TestUserCredentialNotFound()
    {
      using (var session = Connect(null)) {
        Assert.IsFalse((bool)session.User(new NullCredential()));
      }
    }

    [Test]
    public void TestUserNull()
    {
      using (var session = Connect(null)) {
        Assert.IsFalse((bool)session.User((string)null));
      }
    }

    [Test, ExpectedException(typeof(PopIncapableException))]
    public void TestUserIncapable()
    {
      using (var session = Connect(null)) {
        server.EnqueueResponse("+OK List of capabilities follows\r\n" +
                               ".\r\n");

        Assert.IsTrue((bool)session.Capa());
        Assert.IsFalse(session.ServerCapabilities.IsCapable(PopCapability.User));

        StringAssert.AreEqualIgnoringCase("CAPA\r\n",
                                          server.DequeueRequest());

        session.HandlesIncapableAsException = true;
        session.User("user");
      }
    }

    [Test]
    public void TestPass()
    {
      using (var session = Connect(null)) {
        server.EnqueueResponse("+OK mrose is a real hoopy frood\r\n");
        server.EnqueueResponse("-ERR maildrop already locked\r\n");

        Assert.IsTrue((bool)session.User("mrose"));
        Assert.IsFalse((bool)session.Pass("secret"));

        StringAssert.AreEqualIgnoringCase("USER mrose\r\n",
                                          server.DequeueRequest());
        StringAssert.AreEqualIgnoringCase("PASS secret\r\n",
                                          server.DequeueRequest());

        Assert.AreEqual(PopSessionState.Authorization, session.State);
        Assert.AreEqual(authority, session.Authority);

        server.EnqueueResponse("+OK mrose is a real hoopy frood\r\n");
        server.EnqueueResponse("+OK mrose's maildrop has 2 messages (320 octets)\r\n");

        Assert.IsTrue((bool)session.User("mrose"));
        Assert.IsTrue((bool)session.Pass("secret"));

        StringAssert.AreEqualIgnoringCase("USER mrose\r\n",
                                          server.DequeueRequest());
        StringAssert.AreEqualIgnoringCase("PASS secret\r\n",
                                          server.DequeueRequest());

        Assert.AreEqual(PopSessionState.Transaction, session.State);
        Assert.AreEqual(authority, session.Authority);
      }
    }

    [Test]
    public void TestPassCredentialNotFound()
    {
      using (var session = Connect(null)) {
        server.EnqueueResponse("+OK mrose is a real hoopy frood\r\n");

        Assert.IsTrue((bool)session.User("mrose"));

        Assert.IsFalse((bool)session.Pass(new NullCredential()));
      }
    }

    [Test]
    public void TestPassNull()
    {
      using (var session = Connect(null)) {
        server.EnqueueResponse("+OK mrose is a real hoopy frood\r\n");

        Assert.IsTrue((bool)session.User("mrose"));

        Assert.IsFalse((bool)session.Pass((string)null));
      }
    }

    [Test, ExpectedException(typeof(PopProtocolViolationException))]
    public void TestPassBeforeUserCommandIssued()
    {
      using (var session = Connect(null)) {
        session.Pass("secret");
      }
    }

    [Ignore("can't test"), Test, ExpectedException(typeof(PopIncapableException))]
    public void TestPassIncapable()
    {
      using (var session = Connect(null)) {
        server.EnqueueResponse("+OK List of capabilities follows\r\n" +
                               ".\r\n");

        Assert.IsTrue((bool)session.Capa());
        Assert.IsFalse(session.ServerCapabilities.IsCapable(PopCapability.User));

        StringAssert.AreEqualIgnoringCase("CAPA\r\n",
                                          server.DequeueRequest());

        session.HandlesIncapableAsException = true;
        session.User(credential); // USER must be issued first
        session.Pass(credential); // throwing PopIncapableException is expected, but it throwed while USER command
      }
    }

    [Test]
    public void TestLogin()
    {
      using (var session = Connect(null)) {
        server.EnqueueResponse("-ERR sorry, no mailbox for frated here\r\n");

        Assert.IsFalse((bool)session.Login("mrose", "secret"));

        StringAssert.AreEqualIgnoringCase("USER mrose\r\n",
                                          server.DequeueRequest());

        Assert.AreEqual(PopSessionState.Authorization, session.State);
        Assert.AreEqual(authority, session.Authority);

        server.EnqueueResponse("+OK mrose is a real hoopy frood\r\n");
        server.EnqueueResponse("-ERR maildrop already locked\r\n");

        Assert.IsFalse((bool)session.Login("mrose", "secret"));

        StringAssert.AreEqualIgnoringCase("USER mrose\r\n",
                                          server.DequeueRequest());
        StringAssert.AreEqualIgnoringCase("PASS secret\r\n",
                                          server.DequeueRequest());

        Assert.AreEqual(PopSessionState.Authorization, session.State);
        Assert.AreEqual(authority, session.Authority);

        server.EnqueueResponse("+OK mrose is a real hoopy frood\r\n");
        server.EnqueueResponse("+OK mrose's maildrop has 2 messages (320 octets)\r\n");

        Assert.IsTrue((bool)session.Login("mrose", "secret"));

        StringAssert.AreEqualIgnoringCase("USER mrose\r\n",
                                          server.DequeueRequest());
        StringAssert.AreEqualIgnoringCase("PASS secret\r\n",
                                          server.DequeueRequest());

        Assert.AreEqual(PopSessionState.Transaction, session.State);
        Assert.AreEqual(authority, session.Authority);
      }
    }

    [Test]
    public void TestLoginWithNetworkCredential()
    {
      using (var session = Connect(null)) {
        server.EnqueueResponse("+OK\r\n");
        server.EnqueueResponse("+OK 2 messages (320 octets)\r\n");

        Assert.IsTrue((bool)session.Login(credential));

        StringAssert.AreEqualIgnoringCase(string.Format("USER {0}\r\n", credential.UserName),
                                          server.DequeueRequest());
        StringAssert.AreEqualIgnoringCase(string.Format("PASS {0}\r\n", credential.Password),
                                          server.DequeueRequest());

        Assert.AreEqual(PopSessionState.Transaction, session.State);
        Assert.AreEqual(authority, session.Authority);
      }
    }

    [Test]
    public void TestLoginCredentialNotFound()
    {
      using (var session = Connect(null)) {
        Assert.IsFalse((bool)session.Login(new NullCredential()));
        Assert.IsFalse((bool)session.Login(new NullCredential(), "user"));
      }
    }

    [Test]
    public void TestLoginCredentialUsernameNull()
    {
      using (var session = Connect(null)) {
        Assert.IsFalse((bool)session.Login(new NetworkCredential(null, "pass")));
      }
    }

    [Test]
    public void TestLoginCredentialPasswordNull()
    {
      using (var session = Connect(null)) {
        server.EnqueueResponse("+OK\r\n");

        Assert.IsFalse((bool)session.Login(new NetworkCredential("user", null)));

        StringAssert.AreEqualIgnoringCase("USER user\r\n",
                                          server.DequeueRequest());
      }
    }

    [Test, ExpectedException(typeof(PopIncapableException))]
    public void TestApopIncapable()
    {
      using (var session = Connect(null)) {
        session.HandlesIncapableAsException = true;

        Assert.IsFalse(session.ApopAvailable);

        session.Apop(credential, username);
      }
    }

    [Test]
    public void TestApop()
    {
      server.EnqueueResponse("+OK POP3 server ready <1896.697170952@dbc.mtview.ca.us>\r\n");

      using (var session = new PopSession(host, port)) {
        Assert.IsTrue(session.ApopAvailable);

        server.EnqueueResponse("+OK maildrop has 1 message (369 octets)\r\n");

        Assert.IsTrue((bool)session.Apop(new NetworkCredential("mrose", "tanstaaf")));

        StringAssert.AreEqualIgnoringCase("APOP mrose c4c9334bac560ecc979e58001b3e22fb\r\n",
                                          server.DequeueRequest());

        Assert.AreEqual(PopSessionState.Transaction, session.State);
        Assert.AreEqual(new Uri(string.Format("pop://mrose;AUTH=+APOP@{0}:{1}/", host, port)), session.Authority);
      }
    }

    [Test]
    public void TestApopNoWithInUseResponseCode()
    {
      server.EnqueueResponse("+OK POP3 server ready <1896.697170952@dbc.mtview.ca.us>\r\n");

      using (var session = new PopSession(host, port)) {
        Assert.IsTrue(session.ApopAvailable);

        server.EnqueueResponse("-ERR [IN-USE] Do you have another POP session running?\r\n");

        var result = session.Apop(new NetworkCredential("mrose", "tanstaaf"));

        Assert.IsTrue(result.Failed);
        Assert.IsNotNull(result.GetResponseCode(PopResponseCode.InUse));

        StringAssert.AreEqualIgnoringCase("APOP mrose c4c9334bac560ecc979e58001b3e22fb\r\n",
                                          server.DequeueRequest());

        Assert.AreEqual(PopSessionState.Authorization, session.State);
        Assert.AreEqual(new Uri(string.Format("pop://{0}:{1}/", host, port)), session.Authority);
      }
    }

    [Test]
    public void TestApopCredentialNotFound()
    {
      server.EnqueueResponse("+OK POP3 server ready <1896.697170952@dbc.mtview.ca.us>\r\n");

      using (var session = new PopSession(host, port)) {
        Assert.IsTrue(session.ApopAvailable);

        Assert.IsFalse((bool)session.Apop(new NullCredential()));
        Assert.IsFalse((bool)session.Apop(new NullCredential(),"user"));
      }
    }

    [Test]
    public void TestApopCredentialUsernameNull()
    {
      ApopCredentialPropertyNull(new NetworkCredential(null, "password"));
    }

    [Test]
    public void TestApopCredentialPasswordNull()
    {
      ApopCredentialPropertyNull(new NetworkCredential("username", null));
    }

    private void ApopCredentialPropertyNull(ICredentialsByHost credential)
    {
      server.EnqueueResponse("+OK POP3 server ready <1896.697170952@dbc.mtview.ca.us>\r\n");

      using (var session = new PopSession(host, port)) {
        Assert.IsFalse((bool)session.Apop(credential));
      }
    }

    [Test]
    public void TestAuth()
    {
      using (var session = Connect(null)) {
        server.EnqueueResponse("+\r\n");
        server.EnqueueResponse("+OK Maildrop locked and ready\r\n");

        Assert.IsTrue((bool)session.Auth(new NetworkCredential("test", "test", "test"),
                                         PopAuthenticationMechanism.Plain));

        StringAssert.AreEqualIgnoringCase("AUTH PLAIN dGVzdAB0ZXN0AHRlc3Q=\r\n",
                                          server.DequeueRequest());

        Assert.AreEqual(PopSessionState.Transaction, session.State);
        Assert.AreEqual(new Uri(string.Format("pop://test;AUTH=PLAIN@{0}:{1}/", host, port)), session.Authority);
      }
    }

    [Test, ExpectedException(typeof(PopIncapableException))]
    public void TestAuthIncapable()
    {
      using (var session = Connect(null)) {
        server.EnqueueResponse("+OK List of capabilities follows\r\n" +
                               ".\r\n");

        Assert.IsTrue((bool)session.Capa());
        Assert.IsFalse(session.ServerCapabilities.IsCapable(PopCapability.Sasl));

        StringAssert.AreEqualIgnoringCase("CAPA\r\n",
                                          server.DequeueRequest());

        session.HandlesIncapableAsException = true;
        session.Auth(credential, PopAuthenticationMechanism.DigestMD5);
      }
    }

    [Test, ExpectedException(typeof(PopIncapableException))]
    public void TestAuthIncapableMechanism()
    {
      using (var session = Connect(null)) {
        server.EnqueueResponse("+OK List of capabilities follows\r\n" +
                               "SASL PLAIN DIGEST-MD5\r\n" + 
                               ".\r\n");

        Assert.IsTrue((bool)session.Capa());
        Assert.IsFalse(session.ServerCapabilities.IsCapable(PopAuthenticationMechanism.CRAMMD5));

        StringAssert.AreEqualIgnoringCase("CAPA\r\n",
                                          server.DequeueRequest());

        server.EnqueueResponse("-ERR\r\n");

        session.HandlesIncapableAsException = true;
        session.Auth(credential, PopAuthenticationMechanism.CRAMMD5);
      }
    }

    [Test]
    public void TestAuthInvalidResponse()
    {
      using (var session = Connect(null)) {
        server.EnqueueResponse("+ xxxxx-invalid-response-xxxxx\r\n");

        try {
          session.Auth(new NetworkCredential("test", "test", "test"),
                       PopAuthenticationMechanism.DigestMD5);
          Assert.Fail("PopException not thrown");
        }
        catch (PopException) {
        }

        Assert.AreEqual(PopSessionState.NotConnected, session.State);
      }
    }

    [Test]
    public void TestAuthCancelExchanging()
    {
      using (var session = Connect(null)) {
        server.EnqueueResponse("+ eD0xLHk9Mix6PTM=\r\n"); // x=1,y=2,z=3
        server.EnqueueResponse("-ERR AUTH failed.\r\n");

        session.Auth(new NetworkCredential("test", "test", "test"),
                     PopAuthenticationMechanism.DigestMD5);

        StringAssert.AreEqualIgnoringCase("AUTH DIGEST-MD5\r\n", server.DequeueRequest());
        StringAssert.AreEqualIgnoringCase("*\r\n", server.DequeueRequest());
      }
    }

    [Test]
    public void TestAuthCredentialNotFound()
    {
      using (var session = Connect(null)) {
        Assert.IsFalse((bool)session.Auth(new NullCredential(), PopAuthenticationMechanism.DigestMD5));
        Assert.IsFalse((bool)session.Auth(new NullCredential(),"user", PopAuthenticationMechanism.DigestMD5));
      }
    }

    [Test]
    public void TestAuthCredentialUsernameNullInitialClientResponse()
    {
      AuthCredentialPropertyNullInitialClientResponse(new NetworkCredential(null, "password"));
    }

    [Test]
    public void TestAuthCredentialPasswordNullInitialClientResponse()
    {
      AuthCredentialPropertyNullInitialClientResponse(new NetworkCredential("username", null));
    }

    private void AuthCredentialPropertyNullInitialClientResponse(ICredentialsByHost credential)
    {
      using (var session = Connect(null)) {
        Assert.IsFalse((bool)session.Auth(credential,
                                          PopAuthenticationMechanism.Plain));
      }
    }

    [Test]
    public void TestAuthCredentialUsernameNull()
    {
      AuthCredentialPropertyNull(new NetworkCredential(null, "password"));
    }

    [Test]
    public void TestAuthCredentialPasswordNull()
    {
      AuthCredentialPropertyNull(new NetworkCredential("username", null));
    }

    private void AuthCredentialPropertyNull(ICredentialsByHost credential)
    {
      using (var session = Connect(null)) {
        server.EnqueueResponse("+ PDE4OTYuNjk3MTcwOTUyQHBvc3RvZmZpY2UucmVzdG9uLm1jaS5uZXQ+\r\n");
        server.EnqueueResponse("-ERR\r\n");

        Assert.IsFalse((bool)session.Auth(credential,
                                          PopAuthenticationMechanism.CRAMMD5));

        StringAssert.AreEqualIgnoringCase("AUTH CRAM-MD5\r\n", server.DequeueRequest());
        StringAssert.AreEqualIgnoringCase("*\r\n", server.DequeueRequest());
      }
    }

    [Test]
    public void TestAuthNoWithAuthResponseCode()
    {
      using (var session = Connect(null)) {
        server.EnqueueResponse("-ERR [AUTH] PLAIN authentication is disabled\r\n");

        session.HandlesIncapableAsException = false;

        var result = session.Auth(new NetworkCredential("test", "test", "test"),
                                  PopAuthenticationMechanism.Plain);

        Assert.IsTrue(result.Failed);
        Assert.IsNotNull(result.GetResponseCode(PopResponseCode.Auth));

        Assert.AreEqual(PopSessionState.Authorization, session.State);
        Assert.AreEqual(new Uri(string.Format("pop://{0}:{1}/", host, port)), session.Authority);
      }
    }

    [Test]
    public void TestCapa()
    {
      using (var session = Connect(null)) {
        Assert.IsNotNull(session.ServerCapabilities);
        Assert.AreEqual(0, session.ServerCapabilities.Count);

        server.EnqueueResponse("+OK Capability list follows\r\n" +
                               "TOP\r\n" +
                               "USER\r\n" +
                               "SASL CRAM-MD5 KERBEROS_V4\r\n" +
                               "RESP-CODES\r\n" +
                               "LOGIN-DELAY 900\r\n" +
                               "PIPELINING\r\n" +
                               "EXPIRE 60\r\n" +
                               "UIDL\r\n" +
                               "IMPLEMENTATION Shlemazle-Plotz-v302\r\n" +
                               ".\r\n");

        PopCapabilityList capabilities;

        Assert.IsTrue((bool)session.Capa(out capabilities));

        StringAssert.AreEqualIgnoringCase("CAPA\r\n",
                                          server.DequeueRequest());

        Assert.AreEqual(9, capabilities.Count);
        Assert.IsTrue(capabilities.Has(PopCapability.Top));
        Assert.IsTrue(capabilities.Has(PopCapability.User));
        Assert.IsTrue(capabilities.Has(new PopCapability("SASL", "CRAM-MD5", "KERBEROS_V4")));
        Assert.IsTrue(capabilities.Has(PopCapability.RespCodes));
        Assert.IsTrue(capabilities.Has(new PopCapability((string)PopCapability.LoginDelay, "900")));
        Assert.IsTrue(capabilities.Has(PopCapability.Pipelining));
        Assert.IsTrue(capabilities.Has(new PopCapability((string)PopCapability.Expire, "60")));
        Assert.IsTrue(capabilities.Has(PopCapability.Uidl));
        Assert.IsTrue(capabilities.Has(new PopCapability((string)PopCapability.Implementation, "Shlemazle-Plotz-v302")));

        Assert.IsTrue(capabilities.IsCapable(PopCapability.Sasl));
        Assert.IsTrue(capabilities.IsCapable(PopCapability.LoginDelay));
        Assert.IsTrue(capabilities.IsCapable(PopCapability.Expire));
        Assert.IsTrue(capabilities.IsCapable(PopCapability.Implementation));
        Assert.IsTrue(capabilities.IsCapable(PopAuthenticationMechanism.CRAMMD5));
        Assert.IsTrue(capabilities.IsCapable(PopAuthenticationMechanism.KerberosV4));

        Assert.IsNotNull(session.ServerCapabilities);
        Assert.AreEqual(9, session.ServerCapabilities.Count);
        Assert.IsTrue(session.ServerCapabilities.Has(PopCapability.User));
        Assert.IsTrue(session.ServerCapabilities.IsCapable(PopCapability.Sasl));

        try {
          session.ServerCapabilities.Add(PopCapability.Lang);
          Assert.Fail("NotSupportedException not thrown");
        }
        catch (NotSupportedException) {
        }
      }
    }

    [Test]
    public void TestStls()
    {
      using (var session = Connect(null)) {
        server.EnqueueResponse("+OK Begin TLS negotiation\r\n");

        var prevAuthority = new Uri(session.Authority.ToString());

        Assert.IsFalse(session.IsSecureConnection);

        var streamUpgraded = false;

        Assert.IsTrue((bool)session.Stls(delegate(ConnectionBase connection, Stream baseStream) {
          streamUpgraded = true;
          return baseStream; // TODO: return SslStream
        }));

        Assert.IsTrue(streamUpgraded, "stream upgraded");
        Assert.IsTrue(session.IsSecureConnection, "IsSecureConnection");

        StringAssert.AreEqualIgnoringCase("STLS\r\n",
                                          server.DequeueRequest());

        Assert.AreEqual(prevAuthority, session.Authority);
      }
    }

    [Test]
    public void TestStlsExceptionWhileUpgrading()
    {
      using (var session = Connect(null)) {
        server.EnqueueResponse("+OK Begin TLS negotiation\r\n");

        Assert.IsFalse(session.IsSecureConnection);

        try {
          Assert.IsTrue((bool)session.Stls(delegate(ConnectionBase connection, Stream baseStream) {
            throw new System.Security.Authentication.AuthenticationException();
          }));
          Assert.Fail("PopUpgradeConnectionException not thrown");
        }
        catch (PopUpgradeConnectionException) {
        }

        Assert.IsNull(session.Authority);
        Assert.AreEqual(PopSessionState.NotConnected, session.State);
      }
    }

    [Test]
    public void TestStlsError()
    {
      using (var session = Connect(null)) {
        server.EnqueueResponse("-ERR Command not permitted when TLS active\r\n");

        var prevAuthority = new Uri(session.Authority.ToString());

        Assert.IsFalse(session.IsSecureConnection);

        var streamUpgraded = false;

        Assert.IsFalse((bool)session.Stls(delegate(ConnectionBase connection, Stream baseStream) {
          streamUpgraded = true;
          return baseStream; // TODO: return SslStream
        }));

        Assert.IsFalse(streamUpgraded, "stream upgraded");

        StringAssert.AreEqualIgnoringCase("STLS\r\n",
                                          server.DequeueRequest());

        Assert.AreEqual(prevAuthority, session.Authority);
        Assert.IsFalse(session.IsSecureConnection, "IsSecureConnection");
      }
    }

    [Test, ExpectedException(typeof(PopIncapableException))]
    public void TestStlsIncapable()
    {
      using (var session = Connect(null)) {
        server.EnqueueResponse("+OK List of capabilities follows\r\n" +
                               ".\r\n");

        Assert.IsTrue((bool)session.Capa());
        Assert.IsFalse(session.ServerCapabilities.IsCapable(PopCapability.Stls));

        StringAssert.AreEqualIgnoringCase("CAPA\r\n",
                                          server.DequeueRequest());

        Assert.IsFalse(session.IsSecureConnection);

        var streamUpgraded = false;

        session.HandlesIncapableAsException = true;
        session.Stls(delegate(ConnectionBase connection, Stream baseStream) {
          streamUpgraded = true;
          return baseStream;
        });

        Assert.IsFalse(streamUpgraded, "stream upgraded");
        Assert.IsFalse(session.IsSecureConnection, "IsSecureConnection");
      }
    }

    private PopSession Login(params PopCapability[] capabilities)
    {
      var session = Connect(null);

      if (0 < capabilities.Length) {
        var resp = "+OK Capability list follows\r\n" + string.Join("\r\n", (new PopCapabilityList(capabilities)).ToStringArray()) + "\r\n.\r\n";

        server.EnqueueResponse(resp);

        Assert.IsTrue((bool)session.Capa());

        StringAssert.AreEqualIgnoringCase("CAPA\r\n",
                                          server.DequeueRequest());
      }

      server.EnqueueResponse("+OK\r\n");
      server.EnqueueResponse("+OK\r\n");

      Assert.IsTrue((bool)session.Login(credential));

      StringAssert.AreEqualIgnoringCase(string.Format("USER {0}\r\n", credential.UserName),
                                        server.DequeueRequest());
      StringAssert.AreEqualIgnoringCase(string.Format("PASS {0}\r\n", credential.Password),
                                        server.DequeueRequest());

      Assert.AreEqual(PopSessionState.Transaction, session.State);
      Assert.AreEqual(authority, session.Authority);

      session.HandlesIncapableAsException = true;

      return session;
    }

    [Test]
    public void TestUserAfterAuthenticated()
    {
      using (var session = Login()) {
        Assert.IsTrue((bool)session.User(credential));
        Assert.AreEqual(PopSessionState.Transaction, session.State);
      }
    }

    [Test]
    public void TestPassAfterAuthenticated()
    {
      using (var session = Login()) {
        Assert.IsTrue((bool)session.Pass(credential));
        Assert.AreEqual(PopSessionState.Transaction, session.State);
      }
    }

    [Test]
    public void TestAuthAfterAuthenticated()
    {
      using (var session = Login(new PopCapability("SASL", "DIGEST-MD5"))) {
        Assert.IsTrue((bool)session.Auth(credential, PopAuthenticationMechanism.DigestMD5));
        Assert.AreEqual(PopSessionState.Transaction, session.State);
      }
    }

    [Test]
    public void TestApopAfterAuthenticated()
    {
      using (var session = Login()) {
        Assert.IsTrue((bool)session.Apop(credential));
        Assert.AreEqual(PopSessionState.Transaction, session.State);
      }
    }

    [Test]
    public void TestNoOp()
    {
      using (var session = Login()) {
        server.EnqueueResponse("+OK\r\n");

        Assert.IsTrue((bool)session.NoOp());

        StringAssert.AreEqualIgnoringCase("NOOP\r\n",
                                          server.DequeueRequest());
      }
    }

    [Test]
    public void TestRset()
    {
      using (var session = Login()) {
        server.EnqueueResponse("+OK\r\n");

        Assert.IsTrue((bool)session.Rset());

        StringAssert.AreEqualIgnoringCase("RSET\r\n",
                                          server.DequeueRequest());
      }
    }

    [Test]
    public void TestStat()
    {
      using (var session = Login()) {
        server.EnqueueResponse("+OK 2 320\r\n");

        PopDropListing dropListing;

        Assert.IsTrue((bool)session.Stat(out dropListing));

        StringAssert.AreEqualIgnoringCase("STAT\r\n",
                                          server.DequeueRequest());

        Assert.IsNotNull(dropListing);
        Assert.AreEqual(2L, dropListing.MessageCount);
        Assert.AreEqual(320L, dropListing.SizeInOctets);
      }
    }

    [Test]
    public void TestList()
    {
      using (var session = Login()) {
        server.EnqueueResponse("+OK 2 messages (320 octets)\r\n" +
                               "1 120\r\n" + 
                               "2 200\r\n" + 
                               ".\r\n");

        PopScanListing[] scanListings;

        Assert.IsTrue((bool)session.List(out scanListings));

        StringAssert.AreEqualIgnoringCase("LIST\r\n",
                                          server.DequeueRequest());

        Assert.IsNotNull(scanListings);
        Assert.AreEqual(2, scanListings.Length);
        Assert.AreEqual(1L,   scanListings[0].MessageNumber);
        Assert.AreEqual(120L, scanListings[0].SizeInOctets);
        Assert.AreEqual(2L,   scanListings[1].MessageNumber);
        Assert.AreEqual(200L, scanListings[1].SizeInOctets);
      }
    }

    [Test]
    public void TestListWithMessageNumber()
    {
      using (var session = Login()) {
        server.EnqueueResponse("+OK 2 200\r\n");

        PopScanListing scanListing;

        Assert.IsTrue((bool)session.List(2L, out scanListing));

        StringAssert.AreEqualIgnoringCase("LIST 2\r\n",
                                          server.DequeueRequest());

        Assert.IsNotNull(scanListing);
        Assert.AreEqual(2L,   scanListing.MessageNumber);
        Assert.AreEqual(200L, scanListing.SizeInOctets);

        server.EnqueueResponse("-ERR no such message, only 2 messages in maildrop\r\n");

        Assert.IsFalse((bool)session.List(3L, out scanListing));

        StringAssert.AreEqualIgnoringCase("LIST 3\r\n",
                                          server.DequeueRequest());

        Assert.IsNull(scanListing);
      }
    }

    [Test]
    public void TestUidl()
    {
      using (var session = Login(PopCapability.Uidl)) {
        server.EnqueueResponse("+OK\r\n" +
                               "1 whqtswO00WBw418f9t5JxYwZ\r\n" + 
                               "2 QhdPYR:00WBw1Ph7x7\r\n" + 
                               ".\r\n");

        PopUniqueIdListing[] uidListings;

        Assert.IsTrue((bool)session.Uidl(out uidListings));

        StringAssert.AreEqualIgnoringCase("UIDL\r\n",
                                          server.DequeueRequest());

        Assert.IsNotNull(uidListings);
        Assert.AreEqual(2, uidListings.Length);
        Assert.AreEqual(1L, uidListings[0].MessageNumber);
        Assert.AreEqual("whqtswO00WBw418f9t5JxYwZ", uidListings[0].UniqueId);
        Assert.AreEqual(2L, uidListings[1].MessageNumber);
        Assert.AreEqual("QhdPYR:00WBw1Ph7x7", uidListings[1].UniqueId);
      }
    }

    [Test]
    public void TestUidlWithMessageNumber()
    {
      using (var session = Login(PopCapability.Uidl)) {
        server.EnqueueResponse("+OK 2 QhdPYR:00WBw1Ph7x7\r\n");

        PopUniqueIdListing uidListing;

        Assert.IsTrue((bool)session.Uidl(2L, out uidListing));

        StringAssert.AreEqualIgnoringCase("UIDL 2\r\n",
                                          server.DequeueRequest());

        Assert.IsNotNull(uidListing);
        Assert.AreEqual(2L, uidListing.MessageNumber);
        Assert.AreEqual("QhdPYR:00WBw1Ph7x7", uidListing.UniqueId);

        server.EnqueueResponse("-ERR no such message, only 2 messages in maildrop\r\n");

        Assert.IsFalse((bool)session.Uidl(3L, out uidListing));

        StringAssert.AreEqualIgnoringCase("UIDL 3\r\n",
                                          server.DequeueRequest());

        Assert.IsNull(uidListing);
      }
    }

    [Test, ExpectedException(typeof(PopIncapableException))]
    public void TestUidlIncapable()
    {
      using (var session = Login()) {
        PopUniqueIdListing uidListing;

        session.Uidl(1L, out uidListing);
      }
    }

    [Test]
    public void TestDele()
    {
      using (var session = Login()) {
        server.EnqueueResponse("+OK message 1 deleted\r\n");

        Assert.IsTrue((bool)session.Dele(1L));

        StringAssert.AreEqualIgnoringCase("DELE 1\r\n",
                                          server.DequeueRequest());

        server.EnqueueResponse("-ERR no such message, only 2 messages in maildrop\r\n");

        Assert.IsFalse((bool)session.Dele(2L));

        StringAssert.AreEqualIgnoringCase("DELE 2\r\n",
                                          server.DequeueRequest());
      }
    }

    [Test]
    public void TestQuitInTransactionState()
    {
      using (var session = Login()) {
        server.EnqueueResponse("+OK dewey POP3 server signing off (maildrop empty)\r\n");

        Assert.IsTrue((bool)session.Quit());

        StringAssert.AreEqualIgnoringCase("QUIT\r\n",
                                          server.DequeueRequest());

        Assert.AreEqual(PopSessionState.NotConnected, session.State);
        Assert.IsNull(session.Authority);
      }
    }

    [Test]
    public void TestRetr()
    {
      using (var session = Login()) {
        var messageBody = "From: from\r\n" +
          "To: to\r\n" +
          "Subect: subject\r\n" +
          "Date: Mon, 1 Jan 2010 11:53:24 +0900\r\n" + 
          "\r\n" +
          "test message.\r\n";

        server.EnqueueResponse(string.Format("+OK {0} octets\r\n{1}.\r\n", NetworkTransferEncoding.Transfer8Bit.GetByteCount(messageBody), messageBody));

        PopMessage message;

        Assert.IsTrue((bool)session.Retr(1L, out message));

        StringAssert.AreEqualIgnoringCase("RETR 1\r\n",
                                          server.DequeueRequest());

        Assert.IsNotNull(message);
        Assert.AreEqual(1L, message.Number);

        var reader = new StreamReader(message.Body, NetworkTransferEncoding.Transfer8Bit);

        Assert.AreEqual(messageBody, reader.ReadToEnd());

        message.Dispose();


        server.EnqueueResponse("-ERR no such message, only 2 messages in maildrop\r\n");

        Assert.IsFalse((bool)session.Retr(3L, out message));

        StringAssert.AreEqualIgnoringCase("RETR 3\r\n",
                                          server.DequeueRequest());

        Assert.IsNull(message);
      }
    }

    [Test]
    public void TestRetrByteStuffed()
    {
      using (var session = Login()) {
        var messageBody = @"MIME-Version: 1.0
Content-Type: text/plain

..byte-stuffed
+OK response style line
-ERR response style line2
.
end of message
".Replace("\r\n", "\n").Replace("\n", "\r\n");

        server.EnqueueResponse(string.Format("+OK {0} octets\r\n{1}.\r\n",
                                             NetworkTransferEncoding.Transfer8Bit.GetByteCount(messageBody),
                                             messageBody.Replace("\r\n.", "\r\n..")));

        PopMessage message;

        Assert.IsTrue((bool)session.Retr(1L, out message));

        StringAssert.AreEqualIgnoringCase("RETR 1\r\n",
                                          server.DequeueRequest());

        Assert.IsNotNull(message);
        Assert.AreEqual(1L, message.Number);

        var reader = new StreamReader(message.Body, NetworkTransferEncoding.Transfer8Bit);

        Assert.AreEqual(messageBody, reader.ReadToEnd());

        message.Dispose();
      }
    }

    [Test]
    public void TestTop()
    {
      using (var session = Login(PopCapability.Top)) {
        var messageBody = "From: from\r\n" +
          "To: to\r\n" +
          "Subect: subject\r\n" +
          "Date: Mon, 1 Jan 2010 11:53:24 +0900\r\n" + 
          "\r\n";

        server.EnqueueResponse(string.Format("+OK\r\n{0}.\r\n", messageBody));

        PopMessage message;

        Assert.IsTrue((bool)session.Top(1L, 0, out message));

        StringAssert.AreEqualIgnoringCase("TOP 1 0\r\n",
                                          server.DequeueRequest());

        Assert.IsNotNull(message);
        Assert.AreEqual(1L, message.Number);

        var reader = new StreamReader(message.Body, NetworkTransferEncoding.Transfer8Bit);

        Assert.AreEqual(messageBody, reader.ReadToEnd());

        message.Dispose();


        server.EnqueueResponse("-ERR no such message, only 2 messages in maildrop\r\n");

        Assert.IsFalse((bool)session.Top(3L, 10, out message));

        StringAssert.AreEqualIgnoringCase("TOP 3 10\r\n",
                                          server.DequeueRequest());

        Assert.IsNull(message);
      }
    }

    [Test, ExpectedException(typeof(PopIncapableException))]
    public void TestTopIncapable()
    {
      using (var session = Login()) {
        PopMessage message;

        session.Top(1L, 0, out message);
      }
    }

    [Test]
    public void TestGenericCommandWithNoArguments()
    {
      using (var session = Connect(null)) {
        // generic command transaction
        server.EnqueueResponse("+OK done\r\n");

        PopResponse[] responses;

        Assert.IsTrue((bool)session.GenericCommand("X-EXT-COMMAND", out responses));

        StringAssert.AreEqualIgnoringCase("X-EXT-COMMAND\r\n",
                                          server.DequeueRequest());

        Assert.AreEqual(1, responses.Length);
        Assert.IsInstanceOfType(typeof(PopStatusResponse), responses[0]);

        var status = responses[0] as PopStatusResponse;

        Assert.AreEqual("done", status.Text);
      }
    }

    [Test]
    public void TestGenericCommandWithArguments()
    {
      using (var session = Connect(null)) {
        // generic command transaction
        server.EnqueueResponse("+OK done\r\n");

        PopResponse[] responses;

        Assert.IsTrue((bool)session.GenericCommand("X-EXT-COMMAND", out responses, "arg1", "arg2", "arg3"));

        StringAssert.AreEqualIgnoringCase("X-EXT-COMMAND arg1 arg2 arg3\r\n",
                                          server.DequeueRequest());

        Assert.AreEqual(1, responses.Length);
        Assert.IsInstanceOfType(typeof(PopStatusResponse), responses[0]);

        var status = responses[0] as PopStatusResponse;

        Assert.AreEqual("done", status.Text);
      }
    }

    [Test]
    public void TestGenericCommandMultilineResponse()
    {
      using (var session = Connect(null)) {
        // generic command transaction
        server.EnqueueResponse("+OK [X-RESP-CODE] done\r\n" +
                               "1st line\r\n" +
                               "2nd line\r\n" +
                               ".byte-stuffed line\r\n" +
                               ".\r\n");

        PopResponse[] responses;

        Assert.IsTrue((bool)session.GenericCommand("X-EXT-COMMAND", true, out responses));

        StringAssert.AreEqualIgnoringCase("X-EXT-COMMAND\r\n",
                                          server.DequeueRequest());

        Assert.AreEqual(5, responses.Length);
        Assert.IsInstanceOfType(typeof(PopStatusResponse), responses[0]);
        Assert.IsInstanceOfType(typeof(PopFollowingResponse), responses[1]);
        Assert.IsInstanceOfType(typeof(PopFollowingResponse), responses[2]);
        Assert.IsInstanceOfType(typeof(PopFollowingResponse), responses[3]);
        Assert.IsInstanceOfType(typeof(PopTerminationResponse), responses[4]);

        var status = responses[0] as PopStatusResponse;

        Assert.AreEqual("done", status.Text);
        Assert.IsNotNull(status.ResponseText);
        Assert.AreEqual("X-RESP-CODE", (string)status.ResponseText.Code);

        Assert.AreEqual("1st line", (responses[1] as PopFollowingResponse).Text.ToString());
        Assert.AreEqual("2nd line", (responses[2] as PopFollowingResponse).Text.ToString());
        Assert.AreEqual("byte-stuffed line", (responses[3] as PopFollowingResponse).Text.ToString());
      }
    }

    private int port;
    private string host;
    private Uri authority;
    private const string username = "popuser";
    private const string password = "password";
    private NetworkCredential credential;

    private PopPseudoServer server;
  }
}
