using System;
using System.IO;
using System.Net;
using System.Collections.Generic;
using NUnit.Framework;

using Smdn.Net.Pop3.Protocol;
using Smdn.Net.MessageAccessProtocols;

namespace Smdn.Net.Pop3.Client.Session {
  [TestFixture]
  public class PopSessionCreatorTests {
    private class Profile : IPopSessionProfile {
      public Uri Authority {
        get; set;
      }

      public NetworkCredential AnonymousCredential {
        get { return new NetworkCredential("anonymous", "@anonymous"); }
      }

      public int Timeout {
        get; set;
      }

      public int SendTimeout {
        get; set;
      }

      public int ReceiveTimeout {
        get; set;
      }

      public ICredentialsByHost Credentials {
        get; set;
      }

      public string[] UsingSaslMechanisms {
        get; set;
      }

      public bool UseTlsIfAvailable {
        get; set;
      }

      public Profile(NetworkCredential cred, string userinfo, string hostPort)
        : this(cred, PopUri.UriSchemePop, userinfo, hostPort)
      {
      }

      public Profile(NetworkCredential cred, string scheme, string userinfo, string hostPort)
      {
        this.UseTlsIfAvailable = true;
        this.Timeout = 1000;
        this.SendTimeout = -1;
        this.ReceiveTimeout = -1;
        this.Credentials = cred;

        if (userinfo == null)
          this.Authority = new Uri(string.Format("{0}://{1}/", scheme, hostPort));
        else
          this.Authority = new Uri(string.Format("{0}://{1}@{2}/", scheme, userinfo, hostPort));
      }
    }

    [Test, ExpectedException(typeof(ArgumentNullException))]
    public void TestCreateSessionCreateSslStreamCallbackNull()
    {
      using (var server = new PopPseudoServer()) {
        server.Start();

        var prof = new Profile(new NetworkCredential("user", "pass"), "POPS", "user", server.HostPort);

        prof.Timeout = 500;

        server.Stop();

        PopSessionCreator.CreateSession(prof, null);
      }
    }

    [Test, ExpectedException(typeof(PopConnectionException))]
    public void TestCreateSessionConnectFailure()
    {
      using (var server = new PopPseudoServer()) {
        server.Start();

        var prof = new Profile(new NetworkCredential("user", "pass"), "user", server.HostPort);

        prof.Timeout = 500;

        server.Stop();

        PopSessionCreator.CreateSession(prof, null);
      }
    }

    [Test]
    public void TestCreateSessionConnectSecurePort()
    {
      using (var server = new PopPseudoServer()) {
        server.Start();

        var prof = new Profile(new NetworkCredential("user", "pass"), "pops", "user", server.HostPort);

        // greeting
        server.EnqueueResponse("+OK\r\n");
        // CAPA response
        server.EnqueueResponse("+OK\r\n" +
                               "STLS\r\n" +
                               ".\r\n");
        // USER response
        server.EnqueueResponse("+OK\r\n");
        // PASS response
        server.EnqueueResponse("+OK\r\n");

        var streamUpgraded = false;

        using (var session = PopSessionCreator.CreateSession(prof, delegate(ConnectionBase connection, Stream baseStream) {
          streamUpgraded = true;
          return baseStream; // TODO: return SSL stream
        })) {
          Assert.IsTrue(streamUpgraded, "stream upgraded");

          Assert.AreEqual(PopSessionState.Transaction, session.State);
          Assert.AreEqual(prof.Authority, session.Authority);
          Assert.IsTrue(session.IsSecureConnection);
        }
      }
    }

    [Test]
    public void TestCreateSessionUseTlsIfAvailable()
    {
      using (var server = new PopPseudoServer()) {
        server.Start();

        var prof = new Profile(new NetworkCredential("user", "pass"), "user", server.HostPort);

        // greeting
        server.EnqueueResponse("+OK\r\n");
        // CAPA response
        server.EnqueueResponse("+OK\r\n" +
                               "STLS\r\n" +
                               ".\r\n");
        // STLS response
        server.EnqueueResponse("+OK\r\n");
        // CAPA response
        server.EnqueueResponse("+OK\r\n" +
                               ".\r\n");
        // USER response
        server.EnqueueResponse("+OK\r\n");
        // PASS response
        server.EnqueueResponse("+OK\r\n");

        var streamUpgraded = false;

        using (var session = PopSessionCreator.CreateSession(prof, delegate(ConnectionBase connection, Stream baseStream) {
          streamUpgraded = true;
          return baseStream; // TODO: return SSL stream
        })) {
          Assert.IsTrue(streamUpgraded, "stream upgraded");

          Assert.AreEqual(PopSessionState.Transaction, session.State);
          Assert.AreEqual(prof.Authority, session.Authority);
          Assert.IsTrue(session.IsSecureConnection);
        }

        server.DequeueRequest(); // CAPA
        StringAssert.StartsWith("STLS", server.DequeueRequest());
      }
    }

    [Test]
    public void TestCreateSessionStlsFailure()
    {
      using (var server = new PopPseudoServer()) {
        server.Start();

        var prof = new Profile(new NetworkCredential("user", "pass"), "user", server.HostPort);

        // greeting
        server.EnqueueResponse("+OK\r\n");
        // CAPA response
        server.EnqueueResponse("+OK\r\n" +
                               "STLS\r\n" +
                               ".\r\n");
        // STLS response
        server.EnqueueResponse("-ERR\r\n");

        var streamUpgraded = false;

        try {
          using (var session = PopSessionCreator.CreateSession(prof, delegate(ConnectionBase connection, Stream baseStream) {
            streamUpgraded = true;
            return baseStream; // TODO: return SSL stream
          }));
          Assert.Fail("PopUpgradeConnectionException not thrown");
        }
        catch (PopUpgradeConnectionException) {
        }

        Assert.IsFalse(streamUpgraded, "stream upgraded");

        server.DequeueRequest(); // CAPA
        StringAssert.StartsWith("STLS", server.DequeueRequest());
      }
    }

    [Test]
    public void TestAuthenticateUserAndAuthMechanismSpecified()
    {
      using (var server = new PopPseudoServer()) {
        server.Start();

        var prof = new Profile(new NetworkCredential("user", "pass"), "user;auth=plain", server.HostPort);

        // greeting
        server.EnqueueResponse("+OK\r\n");
        // CAPA response
        server.EnqueueResponse("+OK\r\n" +
                               "SASL PLAIN\r\n" +
                               ".\r\n");
        // AUTH response
        server.EnqueueResponse("+ \r\n");
        server.EnqueueResponse("+OK\r\n");

        using (var session = PopSessionCreator.CreateSession(prof, null)) {
          Assert.AreEqual(PopSessionState.Transaction, session.State);
          Assert.AreEqual(prof.Authority, session.Authority);
        }

        server.DequeueRequest(); // CAPA
        StringAssert.StartsWith("AUTH PLAIN", server.DequeueRequest());
      }
    }

    [Test]
    public void TestAuthenticateUserAndAuthMechanismSpecifiedAuthenticationFailure()
    {
      using (var server = new PopPseudoServer()) {
        server.Start();

        var prof = new Profile(new NetworkCredential("user", "pass"), "user;auth=plain", server.HostPort);

        // greeting
        server.EnqueueResponse("+OK\r\n");
        // CAPA response
        server.EnqueueResponse("+OK\r\n" +
                               "SASL PLAIN\r\n" +
                               ".\r\n");
        // AUTH response
        server.EnqueueResponse("+ \r\n");
        server.EnqueueResponse("-ERR\r\n");

        PopSession session = null;

        Assert.IsFalse((bool)PopSessionCreator.CreateSession(prof, null, out session));
        Assert.IsNull(session);

        server.DequeueRequest(); // CAPA
        StringAssert.StartsWith("AUTH PLAIN", server.DequeueRequest());
      }
    }

    [Test]
    public void TestAuthenticateNoUserAndAuthMechanismSpecified()
    {
      using (var server = new PopPseudoServer()) {
        server.Start();

        var prof = new Profile(new NetworkCredential("user", "pass"), ";auth=plain", server.HostPort);

        // greeting
        server.EnqueueResponse("+OK\r\n");
        // CAPA response
        server.EnqueueResponse("+OK\r\n" +
                               "SASL PLAIN\r\n" +
                               ".\r\n");
        // AUTH response
        server.EnqueueResponse("+ \r\n");
        server.EnqueueResponse("+OK\r\n");

        using (var session = PopSessionCreator.CreateSession(prof, null)) {
          Assert.AreEqual(PopSessionState.Transaction, session.State);
          Assert.AreEqual(prof.Authority, session.Authority);
        }

        server.DequeueRequest(); // CAPA
        StringAssert.StartsWith("AUTH PLAIN", server.DequeueRequest());
      }
    }

    [Test]
    public void TestAuthenticateNoUserAndAuthMechanismSpecifiedAuthenticationFailure()
    {
      using (var server = new PopPseudoServer()) {
        server.Start();

        var prof = new Profile(new NetworkCredential("user", "pass"), ";auth=plain", server.HostPort);

        // greeting
        server.EnqueueResponse("+OK\r\n");
        // CAPA response
        server.EnqueueResponse("+OK\r\n" +
                               "SASL PLAIN\r\n" +
                               ".\r\n");
        // AUTH response
        server.EnqueueResponse("+ \r\n");
        server.EnqueueResponse("-ERR\r\n");

        PopSession session = null;

        Assert.IsFalse((bool)PopSessionCreator.CreateSession(prof, null, out session));
        Assert.IsNull(session);

        server.DequeueRequest(); // CAPA
        StringAssert.StartsWith("AUTH PLAIN", server.DequeueRequest());
      }
    }

    [Test]
    public void TestAuthenticateUserAndAppropriateAuthMechanismSpecified()
    {
      AuthenticateWithAppropriateMechanism("user;auth=*");
    }

    [Test]
    public void TestAuthenticateUserAndNoAuthMechanismSpecified()
    {
      AuthenticateWithAppropriateMechanism("user");
    }

    private void AuthenticateWithAppropriateMechanism(string userinfo)
    {
      using (var server = new PopPseudoServer()) {
        server.Start();

        var prof = new Profile(new NetworkCredential("user", "pass"), userinfo, server.HostPort);

        prof.UsingSaslMechanisms = new[] {"LOGIN", "PLAIN"};

        // greeting
        server.EnqueueResponse("+OK\r\n");
        // CAPA response
        server.EnqueueResponse("+OK\r\n" +
                               "SASL DIGEST-MD5 CRAM-MD5 NTLM PLAIN LOGIN\r\n" +
                               ".\r\n");
        // AUTH LOGIN response
        server.EnqueueResponse("+ \r\n");
        server.EnqueueResponse("+OK\r\n");

        using (var session = PopSessionCreator.CreateSession(prof, null)) {
          Assert.AreEqual(PopSessionState.Transaction, session.State);
          Assert.AreEqual(prof.Authority, session.Authority);
        }

        server.DequeueRequest(); // CAPA
        StringAssert.StartsWith("AUTH LOGIN", server.DequeueRequest());
      }
    }

    [Test]
    public void TestAuthenticateUserAndAppropriateAuthMechanismSpecifiedExceptionWhileAuthentication()
    {
      AuthenticateExceptionWhileAuthentication("user;auth=*");
    }

    [Test]
    public void TestAuthenticateUserAndNoAuthMechanismSpecifiedExceptionWhileAuthentication()
    {
      AuthenticateExceptionWhileAuthentication("user");
    }

    private void AuthenticateExceptionWhileAuthentication(string userinfo)
    {
      using (var server = new PopPseudoServer()) {
        server.Start();

        var prof = new Profile(new NetworkCredential("user", "pass"), userinfo, server.HostPort);

        prof.UsingSaslMechanisms = new[] {"DIGEST-MD5", "PLAIN"};

        // greeting
        server.EnqueueResponse("+OK\r\n");
        // CAPA response
        server.EnqueueResponse("+OK\r\n" +
                               "SASL DIGEST-MD5 CRAM-MD5 NTLM PLAIN LOGIN\r\n" +
                               ".\r\n");
        // AUTH DIGEST-MD5 response
        server.EnqueueResponse("+ xxx-invalid-response-xxx\r\n");

        try {
          using (var session = PopSessionCreator.CreateSession(prof, null));
          Assert.Fail("PopException not thrown");
        }
        catch (PopException ex) {
          Assert.IsNotNull(ex.InnerException);
        }
      }
    }

    [Test]
    public void TestAuthenticateUserAndAppropriateAuthMechanismSpecifiedFallbackToNextAppropriateMechanism()
    {
      AuthenticateWithAppropriateMechanismFallbackToNextAppropriateMechanism("user;auth=*");
    }

    [Test]
    public void TestAuthenticateUserAndNoAuthMechanismSpecifiedFallbackToNextAppropriateMechanism()
    {
      AuthenticateWithAppropriateMechanismFallbackToNextAppropriateMechanism("user");
    }

    private void AuthenticateWithAppropriateMechanismFallbackToNextAppropriateMechanism(string userinfo)
    {
      using (var server = new PopPseudoServer()) {
        server.Start();

        var prof = new Profile(new NetworkCredential("user", "pass"), userinfo, server.HostPort);

        prof.UsingSaslMechanisms = new[] {"LOGIN", "PLAIN"};

        // greeting
        server.EnqueueResponse("+OK\r\n");
        // CAPA response
        server.EnqueueResponse("+OK\r\n" +
                               "SASL DIGEST-MD5 CRAM-MD5 NTLM PLAIN LOGIN\r\n" +
                               ".\r\n");
        // AUTH LOGIN response
        server.EnqueueResponse("+ \r\n");
        server.EnqueueResponse("-ERR\r\n");
        // AUTH PLAIN response
        server.EnqueueResponse("+ \r\n");
        server.EnqueueResponse("+OK\r\n");

        using (var session = PopSessionCreator.CreateSession(prof, null)) {
          Assert.AreEqual(PopSessionState.Transaction, session.State);
          Assert.AreEqual(prof.Authority, session.Authority);
        }

        server.DequeueRequest(); // CAPA
        StringAssert.StartsWith("AUTH LOGIN", server.DequeueRequest());
        server.DequeueRequest(); // AUTH LOGIN client response
        StringAssert.StartsWith("AUTH PLAIN", server.DequeueRequest());
      }
    }

    [Test]
    public void TestAuthenticateUserAndAppropriateAuthMechanismSpecifiedFallbackToApop()
    {
      AuthenticateWithAppropriateMechanismFallbackToApop("user;auth=*");
    }

    [Test]
    public void TestAuthenticateUserAndNoAuthMechanismSpecifiedFallbackToApop()
    {
      AuthenticateWithAppropriateMechanismFallbackToApop("user");
    }

    private void AuthenticateWithAppropriateMechanismFallbackToApop(string userinfo)
    {
      using (var server = new PopPseudoServer()) {
        server.Start();

        var prof = new Profile(new NetworkCredential("user", "pass"), userinfo, server.HostPort);

        prof.UsingSaslMechanisms = new[] {"PLAIN", "LOGIN"};

        // greeting
        server.EnqueueResponse("+OK <timestamp@localhost>\r\n");
        // CAPA response
        server.EnqueueResponse("+OK\r\n" +
                               "SASL DIGEST-MD5 CRAM-MD5 NTLM PLAIN LOGIN\r\n" +
                               ".\r\n");
        // AUTH PLAIN response
        server.EnqueueResponse("+ \r\n");
        server.EnqueueResponse("-ERR\r\n");
        // AUTH LOGIN response
        server.EnqueueResponse("+ \r\n");
        server.EnqueueResponse("-ERR\r\n");
        // APOP response
        server.EnqueueResponse("+OK\r\n");

        using (var session = PopSessionCreator.CreateSession(prof, null)) {
          Assert.AreEqual(PopSessionState.Transaction, session.State);
          Assert.AreEqual(prof.Authority, session.Authority);
        }

        server.DequeueRequest(); // CAPA
        StringAssert.StartsWith("AUTH PLAIN", server.DequeueRequest());
        server.DequeueRequest(); // AUTH PLAIN client response
        StringAssert.StartsWith("AUTH LOGIN", server.DequeueRequest());
        server.DequeueRequest(); // AUTH LOGIN client response
        StringAssert.StartsWith("APOP user", server.DequeueRequest());
      }
    }

    [Test]
    public void TestAuthenticateUserAndAppropriateAuthMechanismSpecifiedFallbackToUserPass()
    {
      AuthenticateWithAppropriateMechanismFallbackToUserPass("user;auth=*");
    }

    [Test]
    public void TestAuthenticateUserAndNoAuthMechanismSpecifiedFallbackToUserPass()
    {
      AuthenticateWithAppropriateMechanismFallbackToUserPass("user");
    }

    private void AuthenticateWithAppropriateMechanismFallbackToUserPass(string userinfo)
    {
      using (var server = new PopPseudoServer()) {
        server.Start();

        var prof = new Profile(new NetworkCredential("user", "pass"), userinfo, server.HostPort);

        // greeting
        server.EnqueueResponse("+OK\r\n");
        // CAPA response
        server.EnqueueResponse("+OK\r\n" +
                               ".\r\n");
        // USER response
        server.EnqueueResponse("+OK\r\n");
        // PASS response
        server.EnqueueResponse("+OK\r\n");

        using (var session = PopSessionCreator.CreateSession(prof, null)) {
          Assert.AreEqual(PopSessionState.Transaction, session.State);
          Assert.AreEqual(prof.Authority, session.Authority);
        }

        server.DequeueRequest(); // CAPA
        StringAssert.StartsWith("USER user", server.DequeueRequest());
        StringAssert.StartsWith("PASS pass", server.DequeueRequest());
      }
    }

    [Test]
    public void TestAuthenticateUserAndAppropriateAuthMechanismSpecifiedFallbackToUserPassAuthenticationFailure()
    {
      AuthenticateWithAppropriateMechanismFallbackToUserPassAuthenticationFailure("user;auth=*");
    }

    [Test]
    public void TestAuthenticateUserAndNoAuthMechanismSpecifiedFallbackToUserPassAuthenticationFailure()
    {
      AuthenticateWithAppropriateMechanismFallbackToUserPassAuthenticationFailure("user");
    }

    private void AuthenticateWithAppropriateMechanismFallbackToUserPassAuthenticationFailure(string userinfo)
    {
      using (var server = new PopPseudoServer()) {
        server.Start();

        var prof = new Profile(new NetworkCredential("user", "pass"), userinfo, server.HostPort);

        // greeting
        server.EnqueueResponse("+OK\r\n");
        // CAPA response
        server.EnqueueResponse("+OK\r\n" +
                               ".\r\n");
        // USER response
        server.EnqueueResponse("-ERR\r\n");

        PopSession session = null;

        Assert.IsFalse((bool)PopSessionCreator.CreateSession(prof, null, out session));
        Assert.IsNull(session);

        server.DequeueRequest(); // CAPA
        StringAssert.StartsWith("USER user", server.DequeueRequest());
      }
    }

    [Test, Ignore("not work; LOGINDISABLE")]
    public void TestAuthenticateUserAndAppropriateAuthMechanismSpecifiedAppropriateMechanismNotFound()
    {
      AuthenticateWithAppropriateMechanismAppropriateMechanismNotFound("user;auth=*");
    }

    [Test, Ignore("not work; LOGINDISABLE")]
    public void TestAuthenticateUserAndNoAuthMechanismSpecifiedAppropriateMechanismNotFound()
    {
      AuthenticateWithAppropriateMechanismAppropriateMechanismNotFound("user");
    }

    private void AuthenticateWithAppropriateMechanismAppropriateMechanismNotFound(string userinfo)
    {
      using (var server = new PopPseudoServer()) {
        server.Start();

        var prof = new Profile(new NetworkCredential("user", "pass"), userinfo, server.HostPort);

        // greeting
        server.EnqueueResponse("+OK\r\n");
        // CAPA response
        server.EnqueueResponse("+OK\r\n" +
                               ".\r\n");
        // TODO: LOGINDISABLE

        try {
          using (var session = PopSessionCreator.CreateSession(prof, null));
          Assert.Fail("PopException not thrown");
        }
        catch (PopException ex) {
          Assert.IsNotNull(ex.InnerException);
        }

        server.DequeueRequest(); // CAPA
      }
    }

    [Test]
    public void TestAuthenticateAnonymousMechanismSpecified()
    {
      using (var server = new PopPseudoServer()) {
        server.Start();

        var prof = new Profile(new NetworkCredential("user", "pass"), "user;auth=anonymous", server.HostPort);

        // greeting
        server.EnqueueResponse("+OK\r\n");
        // CAPA response
        server.EnqueueResponse("+OK\r\n" +
                               "SASL ANONYMOUS\r\n" +
                               ".\r\n");
        // AUTH response
        server.EnqueueResponse("+ \r\n");
        server.EnqueueResponse("+OK\r\n");

        using (var session = PopSessionCreator.CreateSession(prof, null)) {
          Assert.AreEqual(PopSessionState.Transaction, session.State);
          Assert.AreEqual(prof.Authority, session.Authority);
        }

        server.DequeueRequest(); // CAPA
        StringAssert.StartsWith("AUTH ANONYMOUS", server.DequeueRequest());
      }
    }

    [Test]
    public void TestAuthenticateAnonymousMechanismSpecifiedAuthenticationFailure()
    {
      using (var server = new PopPseudoServer()) {
        server.Start();

        var prof = new Profile(new NetworkCredential("user", "pass"), "user;auth=anonymous", server.HostPort);

        // greeting
        server.EnqueueResponse("+OK\r\n");
        // CAPA response
        server.EnqueueResponse("+OK\r\n" +
                               "SASL ANONYMOUS\r\n" +
                               ".\r\n");
        // AUTH response
        server.EnqueueResponse("+ \r\n");
        server.EnqueueResponse("-ERR\r\n");

        PopSession session = null;

        Assert.IsFalse((bool)PopSessionCreator.CreateSession(prof, null, out session));
        Assert.IsNull(session);

        server.DequeueRequest(); // CAPA
        StringAssert.StartsWith("AUTH ANONYMOUS", server.DequeueRequest());
      }
    }

    [Test]
    public void TestAuthenticateNoUserAndNoAuthMechanismSpecified()
    {
      using (var server = new PopPseudoServer()) {
        server.Start();

        var prof = new Profile(new NetworkCredential("user", "pass"), null, server.HostPort);

        // greeting
        server.EnqueueResponse("+OK\r\n");
        // CAPA response
        server.EnqueueResponse("+OK\r\n" +
                               "SASL ANONYMOUS\r\n" +
                               ".\r\n");
        // AUTH response
        server.EnqueueResponse("+ \r\n");
        server.EnqueueResponse("-ERR\r\n");
        // USER response
        server.EnqueueResponse("+OK\r\n");
        // PASS response
        server.EnqueueResponse("+OK\r\n");

        using (var session = PopSessionCreator.CreateSession(prof, null)) {
          Assert.AreEqual(PopSessionState.Transaction, session.State);
          Assert.AreEqual(prof.Authority, session.Authority);
        }

        server.DequeueRequest(); // CAPA
        StringAssert.StartsWith("AUTH ANONYMOUS", server.DequeueRequest());
        server.DequeueRequest(); // AUTH ANONYMOUS client response
        StringAssert.StartsWith("USER anonymous", server.DequeueRequest()); // anonymous credential
        StringAssert.StartsWith("PASS", server.DequeueRequest());
      }
    }

    [Test]
    public void TestAuthenticateNoUserAndNoAuthMechanismSpecifiedAuthenticationFailure()
    {
      using (var server = new PopPseudoServer()) {
        server.Start();

        var prof = new Profile(new NetworkCredential("user", "pass"), null, server.HostPort);

        // greeting
        server.EnqueueResponse("+OK\r\n");
        // CAPA response
        server.EnqueueResponse("+OK\r\n" +
                               ".\r\n");
        // USER response
        server.EnqueueResponse("+OK\r\n");
        // PASS response
        server.EnqueueResponse("-ERR\r\n");

        PopSession session = null;

        Assert.IsFalse((bool)PopSessionCreator.CreateSession(prof, null, out session));
        Assert.IsNull(session);

        server.DequeueRequest(); // CAPA
        StringAssert.StartsWith("USER anonymous", server.DequeueRequest()); // anonymous credential
        StringAssert.StartsWith("PASS", server.DequeueRequest());
      }
    }

    [Test, Ignore("not work; LOGINDISABLE")]
    public void TestAuthenticateNoUserAndNoAuthMechanismSpecifiedAppropriateMechanismNotFound()
    {
      using (var server = new PopPseudoServer()) {
        server.Start();

        var prof = new Profile(new NetworkCredential("user", "pass"), null, server.HostPort);

        // greeting
        server.EnqueueResponse("+OK\r\n");
        // CAPA response
        server.EnqueueResponse("+OK\r\n" +
                               ".\r\n");
        // TODO: LOGINDISABLE

        try {
          using (var session = PopSessionCreator.CreateSession(prof, null));
          Assert.Fail("PopException not thrown");
        }
        catch (PopException ex) {
          Assert.IsNull(ex.InnerException);
        }

        server.DequeueRequest(); // CAPA
      }
    }
  }
}
