// 
// Author:
//       smdn <smdn@mail.invisiblefulmoon.net>
// 
// Copyright (c) 2008-2010 smdn
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.Net;
using System.Reflection;

namespace Smdn.Security.Authentication.Sasl.Client {
  public abstract class SaslClientMechanism : IDisposable {
    /*
     * class members
     */
    static SaslClientMechanism()
    {
      var type = typeof(SaslClientMechanism);

      foreach (var t in Assembly.GetExecutingAssembly().GetTypes()) {
        if (t.IsAbstract || !type.IsAssignableFrom(t))
          continue;

        RegisterMechanism(t);
      }
    }

    public static SaslClientMechanism Create(string mechanismName)
    {
      if (string.IsNullOrEmpty(mechanismName))
        throw new ArgumentException("invalid mechanism name", mechanismName);

      try {
        if (knownMechanismTypes.ContainsKey(mechanismName))
          return (SaslClientMechanism)Activator.CreateInstance(knownMechanismTypes[mechanismName]);
        else
          throw new SaslMechanismNotSupportedException(mechanismName);
      }
      catch (TargetInvocationException ex) {
        if (ex.InnerException is NotSupportedException)
          throw new SaslMechanismNotSupportedException(mechanismName, ex.InnerException);
        else
          throw new SaslException("can't create instance", ex);
      }
    }

    public static string[] GetAvailableMechanisms()
    {
      return (new List<string>(knownMechanismTypes.Keys)).ToArray();
    }

    public static void RegisterMechanism(Type mechanismType)
    {
      if (mechanismType == null)
        throw new ArgumentNullException("mechanismType");
      else if (mechanismType.IsAbstract)
        throw new ArgumentException("can't register abstract class type", "mechanismType");
      else if (!typeof(SaslClientMechanism).IsAssignableFrom(mechanismType))
        throw new ArgumentException("type is not SaslClientMechanism", "mechanismType");

      var attrs = mechanismType.GetCustomAttributes(typeof(SaslMechanismNameAttribute), false);

      if (attrs == null || attrs.Length <= 0)
        throw new ArgumentException("SaslMechanismNameAttribute not applied", "mechanismType");

      var name = (attrs[0] as SaslMechanismNameAttribute).Name;

      if (string.IsNullOrEmpty(name))
        throw new ArgumentException("invalid mechanism name", "mechanismType");

      lock (((System.Collections.ICollection)knownMechanismTypes).SyncRoot) {
        knownMechanismTypes[name] = mechanismType;
      }
    }

    private static Dictionary<string, Type> knownMechanismTypes = new Dictionary<string, Type>(StringComparer.InvariantCultureIgnoreCase);

    /*
     * instance members
     */
    public virtual bool ClientFirst  {
      get { return false; }
    }

    public NetworkCredential Credential {
      get { CheckDisposed(); return credential; }
      set { CheckDisposed(); credential = value; }
    }

    /// <value>see http://www.iana.org/assignments/gssapi-service-names</value>
    public string ServiceName {
      get; set;
    }

    protected SaslClientMechanism()
    {
      Initialize();
    }

    ~SaslClientMechanism()
    {
      Dispose(false);
    }

    public void Dispose()
    {
      Dispose(true);
      GC.SuppressFinalize(this);
    }

    protected virtual void Dispose(bool disposing)
    {
      disposed = true;

      credential = null;
    }

    public virtual void Initialize()
    {
      CheckDisposed();
    }

    public virtual SaslExchangeStatus GetInitialResponse(out byte[] initialClientResponse)
    {
      if (ClientFirst)
        return Exchange(null, out initialClientResponse);
      else
        throw new NotSupportedException("initial response is not supported by this mechanism");
    }

    public SaslExchangeStatus Exchange(byte[] serverChallenge, out byte[] clientResponse)
    {
      CheckDisposed();

      ByteString resp;

      var status = Exchange((serverChallenge == null) ? ByteString.CreateEmpty() : new ByteString(serverChallenge), out resp);

      clientResponse = (resp == null) ? null : resp.ByteArray;

      return status;
    }

    protected abstract SaslExchangeStatus Exchange(ByteString serverChallenge, out ByteString clientResponse);

    protected void CheckDisposed()
    {
      if (disposed)
        throw new ObjectDisposedException(GetType().FullName);
    }

    private bool disposed = false;
    private NetworkCredential credential = null;
  }
}
