using System;
using System.Net;
using System.Text;
using NUnit.Framework;

namespace Smdn.Security.Authentication.Sasl.Client {
  [TestFixture]
  public class NTLMMechanismTests {
    [Test, ExpectedException(typeof(SaslException))]
    public void TestCredentialNotSet()
    {
      using (var client = new NTLMMechanism()) {
        byte[] clientResponse;

        client.Exchange(new byte[0], out clientResponse);
      }
    }

    [Test]
    public void TestExchange()
    {
      /*
       * http://davenport.sourceforge.net/ntlm.html
       * Appendix B: Application Protocol Usage of NTLM - NTLM IMAP Authentication
       */
      using (var client = new NTLMMechanism()) {
        client.Credential = new NetworkCredential("user", "SecREt01", "DOMAIN");
        client.TargetHost = "WORKSTATION";

        /*
         * Type-1
         */
        byte[] serverChallenge;
        byte[] clientResponse;

        serverChallenge = new byte[0];

        Assert.AreEqual(SaslExchangeStatus.Continuing,
                        client.Exchange(serverChallenge, out clientResponse));

        BytesAssert.AreEqual(Convert.FromBase64String("TlRMTVNTUAABAAAABzIAAAYABgArAAAACwALACAAAABXT1JLU1RBVElPTkRPTUFJTg=="),
                             clientResponse);

        /*
         * Type-3
         */
        serverChallenge = Convert.FromBase64String("TlRMTVNTUAACAAAADAAMADAAAAABAoEAASNFZ4mrze8AAAAAAAAAAGIAYgA8AAAA" +
                                                   "RABPAE0AQQBJAE4AAgAMAEQATwBNAEEASQBOAAEADABTAEUAUgBWAEUAUgAEABQAZA" +
                                                   "BvAG0AYQBpAG4ALgBjAG8AbQADACIAcwBlAHIAdgBlAHIALgBkAG8AbQBhAGkAbgAu" +
                                                   "AGMAbwBtAAAAAAA=");

        Assert.AreEqual(SaslExchangeStatus.Succeeded,
                        client.Exchange(serverChallenge, out clientResponse));

        BytesAssert.AreEqual(Convert.FromBase64String("TlRMTVNTUAADAAAAGAAYAGoAAAAYABgAggAAAAwADABAAAAACAAIAEwAAAAWABYAVA" +
                                                      "AAAAAAAACaAAAAAQIAAEQATwBNAEEASQBOAHUAcwBlAHIAVwBPAFIASwBTAFQAQQBU" +
                                                      "AEkATwBOAMM3zVy9RPyXgqZnr21CfG3mfCDC0+d8ViWpjBwx6BhHRmspst9GgPOZWP" +
                                                      "uMITqcxg=="),
                             clientResponse);
      }
    }

    [Test]
    public void TestExchangeCredentialUsernameEmpty()
    {
      ExchangeEmptyCredentialProperty(new NetworkCredential(null, "pass", "DOMAIN"));
    }

    [Test]
    public void TestExchangeCredentialPaswordEmpty()
    {
      ExchangeEmptyCredentialProperty(new NetworkCredential("user", null, "DOMAIN"));
    }

    private void ExchangeEmptyCredentialProperty(NetworkCredential credential)
    {
      using (var client = new NTLMMechanism()) {
        client.Credential = credential;
        client.TargetHost = "WORKSTATION";

        /*
         * Type-1
         */
        byte[] serverChallenge;
        byte[] clientResponse;

        serverChallenge = new byte[0];

        Assert.AreEqual(SaslExchangeStatus.Continuing,
                        client.Exchange(serverChallenge, out clientResponse));

        BytesAssert.AreEqual(Convert.FromBase64String("TlRMTVNTUAABAAAABzIAAAYABgArAAAACwALACAAAABXT1JLU1RBVElPTkRPTUFJTg=="),
                             clientResponse);

        /*
         * Type-3
         */
        serverChallenge = Convert.FromBase64String("TlRMTVNTUAACAAAADAAMADAAAAABAoEAASNFZ4mrze8AAAAAAAAAAGIAYgA8AAAA" +
                                                   "RABPAE0AQQBJAE4AAgAMAEQATwBNAEEASQBOAAEADABTAEUAUgBWAEUAUgAEABQAZA" +
                                                   "BvAG0AYQBpAG4ALgBjAG8AbQADACIAcwBlAHIAdgBlAHIALgBkAG8AbQBhAGkAbgAu" +
                                                   "AGMAbwBtAAAAAAA=");

        Assert.AreEqual(SaslExchangeStatus.Failed,
                        client.Exchange(serverChallenge, out clientResponse));
        Assert.IsNull(clientResponse);
      }
    }

    [Test]
    public void TestGetInitialResponse()
    {
      using (var client = new NTLMMechanism()) {
        Assert.IsTrue(client.ClientFirst);

        client.Credential = new NetworkCredential("user", "pass", "DOMAIN");
        client.TargetHost = "WORKSTATION";

        byte[] initialResponse;

        Assert.AreNotEqual(SaslExchangeStatus.Failed,
                           client.GetInitialResponse(out initialResponse));
      }
    }
  }
}
