using System;
using System.Net;
using NUnit.Framework;

namespace Smdn.Security.Authentication.Sasl.Client {
  [TestFixture]
  public class PlainMechanismTests {
    [Test, ExpectedException(typeof(SaslException))]
    public void TestCredentialNotSet()
    {
      using (var client = new PlainMechanism()) {
        byte[] clientResponse;

        client.Exchange(new byte[0], out clientResponse);
      }
    }

    [Test]
    public void TestExchangeWithoutAuthcid()
    {
      using (var client = new PlainMechanism()) {
        client.Credential = new NetworkCredential("tim", "tanstaaftanstaaf");

        byte[] clientResponse;

        Assert.AreEqual(SaslExchangeStatus.Succeeded,
                        client.Exchange(new byte[0], out clientResponse));

        BytesAssert.AreEqual("\0tim\0tanstaaftanstaaf", clientResponse);
      }
    }

    [Test]
    public void TestExchangeWithAuthcid()
    {
      using (var client = new PlainMechanism()) {
        client.Credential = new NetworkCredential("Kurt", "xipj3plmq", "Ursel");

        byte[] serverChallenge = new byte[0];
        byte[] clientResponse;

        Assert.AreEqual(SaslExchangeStatus.Succeeded,
                        client.Exchange(serverChallenge, out clientResponse));

        BytesAssert.AreEqual("Ursel\0Kurt\0xipj3plmq", clientResponse);
      }
    }

    [Test]
    public void TestExchangeCredentialUsernameEmpty()
    {
      ExchangeEmptyCredentialProperty(new NetworkCredential(null, "tanstaaftanstaaf"));
    }

    [Test]
    public void TestExchangeCredentialPasswordEmpty()
    {
      ExchangeEmptyCredentialProperty(new NetworkCredential("tim", null));
    }

    private void ExchangeEmptyCredentialProperty(NetworkCredential credential)
    {
      using (var client = new PlainMechanism()) {
        client.Credential = credential;

        byte[] serverChallenge = new byte[0];
        byte[] clientResponse;

        Assert.AreEqual(SaslExchangeStatus.Failed,
                        client.Exchange(serverChallenge, out clientResponse));
        Assert.IsNull(clientResponse);
      }
    }

    [Test]
    public void TestGetInitialResponse()
    {
      using (var client = new PlainMechanism()) {
        Assert.IsTrue(client.ClientFirst);

        client.Credential = new NetworkCredential("test", "test", "test");

        byte[] initialResponse;

        Assert.AreNotEqual(SaslExchangeStatus.Failed,
                           client.GetInitialResponse(out initialResponse));

        BytesAssert.AreEqual(Convert.FromBase64String("dGVzdAB0ZXN0AHRlc3Q="),
                             initialResponse);
      }
    }
  }
}
